import codecs
import dataclasses
import hashlib
import json
import os
from pathlib import Path
from typing import Any, List

from garaga.definitions import CURVES, CurveID, G1Point, G2Point
from garaga.hints import io
from garaga.hints.io import split_128
from garaga.modulo_circuit_structs import (
    E12D,
    G1PointCircuit,
    G2PointCircuit,
    StructArray,
)
from garaga.precompiled_circuits.multi_miller_loop import MultiMillerLoopCircuit

# Import constants from generated file
from garaga.starknet.constants import (
    RISC0_BN254_CONTROL_ID,
    RISC0_CONTROL_ROOT,
    SP1_VERIFIER_HASH,
    SP1_VERIFIER_VERSION,
)

# https://github.com/risc0/risc0-ethereum/blob/main/contracts/src/groth16/ControlID.sol
# release 2.0
RISC0_CONTROL_ROOT = 0x539032186827B06719244873B17B2D4C122E2D02CFB1994FE958B2523B844576
RISC0_BN254_CONTROL_ID = (
    0x04446E66D300EB7FB45C9726BB53C793DDA407A62E9601618BB43C5C14657AC0
)

# https://github.com/succinctlabs/sp1-contracts/blob/main/contracts/src/v4.0.0-rc.3/SP1VerifierGroth16.sol
SP1_VERIFIER_VERSION: str = "v4.0.0-rc.3"
SP1_VERIFIER_HASH: bytes = bytes.fromhex(
    "11b6a09d63d255ad425ee3a7f6211d5ec63fbde9805b40551c3136275b6f4eb4"
)


class KeyPatternNotFound(Exception):
    def __init__(self, key_patterns):
        super().__init__(f"No key found with patterns {key_patterns}")
        self.key_patterns = key_patterns


def iterate_nested_dict(d):
    for key, value in d.items():
        if isinstance(value, dict):
            yield from iterate_nested_dict(value)
        else:
            yield value


def find_item_from_key_patterns(data: dict, key_patterns: List[str]) -> Any:
    best_match = None
    best_score = -1
    for key, value in data.items():
        for pattern in key_patterns:
            if key.lower() == pattern.lower():
                # Exact match
                return value
            elif pattern.lower() in key.lower():
                # Partial match
                score = key.lower().count(pattern.lower())
                if score > best_score:
                    best_match = value
                    best_score = score

    if best_match is not None:
        return best_match
    else:
        raise KeyPatternNotFound(key_patterns)


def try_parse_g1_point_from_key(
    data: dict, key_patterns: List[str], curve_id: CurveID = None
) -> G1Point:
    point = find_item_from_key_patterns(data, key_patterns)
    return try_parse_g1_point(point, curve_id)


def proj_to_affine(x, y, z, curve_id: CurveID) -> G1Point:
    x, y, z = io.to_int(x), io.to_int(y), io.to_int(z)
    p = curve_id.p
    z = pow(z, -1, p)
    x = x * z % p
    y = y * z % p
    return G1Point(x=x, y=y, curve_id=curve_id)


def try_parse_g1_point(point: Any, curve_id: CurveID = None) -> G1Point:
    if isinstance(point, dict):
        return G1Point(
            x=io.to_int(find_item_from_key_patterns(point, ["x"])),
            y=io.to_int(find_item_from_key_patterns(point, ["y"])),
            curve_id=curve_id,
        )
    elif isinstance(point, (tuple, list)):
        if len(point) == 2:
            return G1Point(
                x=io.to_int(point[0]),
                y=io.to_int(point[1]),
                curve_id=curve_id,
            )
        elif len(point) == 3:
            return proj_to_affine(point[0], point[1], point[2], curve_id)
        else:
            raise ValueError(f"Invalid point: {point}")
    else:
        raise ValueError(f"Invalid point: {point}")


def try_parse_g2_point_from_key(
    data: dict, key_patterns: List[str], curve_id: CurveID = None
) -> G2Point:
    point = find_item_from_key_patterns(data, key_patterns)
    return try_parse_g2_point(point, curve_id)


def try_parse_g2_point(point: Any, curve_id: CurveID = None) -> G2Point:
    if isinstance(point, dict):
        x_g2 = find_item_from_key_patterns(point, ["x"])
        y_g2 = find_item_from_key_patterns(point, ["y"])
        if isinstance(x_g2, dict) and isinstance(y_g2, dict):
            return G2Point(
                x=(
                    io.to_int(find_item_from_key_patterns(x_g2, ["a0"])),
                    io.to_int(find_item_from_key_patterns(x_g2, ["a1"])),
                ),
                y=(
                    io.to_int(find_item_from_key_patterns(y_g2, ["a0"])),
                    io.to_int(find_item_from_key_patterns(y_g2, ["a1"])),
                ),
                curve_id=curve_id,
            )
        elif isinstance(x_g2, (tuple, list)) and isinstance(y_g2, (tuple, list)):
            return G2Point(
                x=(
                    io.to_int(x_g2[0]),
                    io.to_int(x_g2[1]),
                ),
                y=(
                    io.to_int(y_g2[0]),
                    io.to_int(y_g2[1]),
                ),
                curve_id=curve_id,
            )
    elif isinstance(point, (tuple, list)):
        if len(point) == 2:
            supposed_x = point[0]
            supposed_y = point[1]
        elif len(point) == 3:
            assert (io.to_int(point[2][0]), io.to_int(point[2][1])) == (
                1,
                0,
            ), "Non standard projective coordinates"
            supposed_x = point[0]
            supposed_y = point[1]

        if isinstance(supposed_x, (tuple, list)):
            assert len(supposed_x) == 2, f"Invalid fp2 coordinates: {supposed_x}"
            supposed_x = (io.to_int(supposed_x[0]), io.to_int(supposed_x[1]))
        if isinstance(supposed_y, (tuple, list)):
            assert len(supposed_y) == 2, f"Invalid fp2 coordinates: {supposed_y}"
            supposed_y = (io.to_int(supposed_y[0]), io.to_int(supposed_y[1]))
        return G2Point(x=supposed_x, y=supposed_y, curve_id=curve_id)
    else:
        raise ValueError(f"Invalid point: {point}")


def try_guessing_curve_id_from_json(data: dict) -> CurveID:
    try:
        curve_id = CurveID.from_str(find_item_from_key_patterns(data, ["curve"]))
    except (ValueError, KeyError, KeyPatternNotFound):
        # Try guessing the curve id from the bit size of the first found integer in the json.
        x = None
        for value in iterate_nested_dict(data):
            try:
                x = io.to_int(value)
                break
            except TypeError:
                continue
        if x is None:
            raise ValueError("No integer found in the JSON data.")
        if x.bit_length() > 256:
            curve_id = CurveID.BLS12_381
        else:
            curve_id = CurveID.BN254
    return curve_id


@dataclasses.dataclass(slots=True)
class Groth16VerifyingKey:
    alpha: G1Point
    beta: G2Point
    gamma: G2Point
    delta: G2Point
    ic: List[G1Point]

    def __post_init__(self):
        assert (
            len(self.ic) > 1
        ), "The verifying key must have at least two points in the ic array (one public input and one private input)"
        assert len(self.ic) == len(
            set(self.ic)
        ), "The ic array must not contain duplicate points."
        assert (
            self.alpha.curve_id
            == self.beta.curve_id
            == self.gamma.curve_id
            == self.delta.curve_id
        ), "All points must be on the same curve."
        assert all(point.curve_id == self.alpha.curve_id for point in self.ic)

    @property
    def curve_id(self) -> CurveID:
        return self.alpha.curve_id

    def from_dict(data: dict) -> "Groth16VerifyingKey":
        try:
            curve_id = try_guessing_curve_id_from_json(data)
            try:
                verifying_key = find_item_from_key_patterns(data, ["verifying_key"])
            except KeyPatternNotFound:
                verifying_key = data
            try:
                return Groth16VerifyingKey(
                    alpha=try_parse_g1_point_from_key(
                        verifying_key, ["alpha"], curve_id
                    ),
                    beta=try_parse_g2_point_from_key(verifying_key, ["beta"], curve_id),
                    gamma=try_parse_g2_point_from_key(
                        verifying_key, ["gamma"], curve_id
                    ),
                    delta=try_parse_g2_point_from_key(
                        verifying_key, ["delta"], curve_id
                    ),
                    ic=[
                        try_parse_g1_point(point, curve_id)
                        for point in find_item_from_key_patterns(verifying_key, ["ic"])
                    ],
                )
            except (ValueError, KeyPatternNotFound):
                # Gnark case.
                g1_points = find_item_from_key_patterns(verifying_key, ["g1"])
                g2_points = find_item_from_key_patterns(verifying_key, ["g2"])
                return Groth16VerifyingKey(
                    alpha=try_parse_g1_point_from_key(g1_points, ["alpha"], curve_id),
                    beta=try_parse_g2_point_from_key(g2_points, ["beta"], curve_id),
                    gamma=try_parse_g2_point_from_key(g2_points, ["gamma"], curve_id),
                    delta=try_parse_g2_point_from_key(g2_points, ["delta"], curve_id),
                    ic=[
                        try_parse_g1_point(point, curve_id)
                        for point in find_item_from_key_patterns(g1_points, ["K"])
                    ],
                )
        except KeyError as e:
            raise KeyError(f"The key {e} is missing from the JSON data.")

    def from_json(file_path: str | Path) -> "Groth16VerifyingKey":
        path = Path(file_path)
        try:
            with path.open("r") as f:
                data = json.load(f)
        except FileNotFoundError:
            cwd = os.getcwd()
            print(f"Current working directory: {cwd}")
            print(f"Attempted to access file: {os.path.abspath(file_path)}")
            raise FileNotFoundError(f"The file {file_path} was not found.")
        except json.JSONDecodeError:
            raise ValueError(f"The file {file_path} does not contain valid JSON.")
        return Groth16VerifyingKey.from_dict(data)

    def serialize_to_cairo(self) -> str:
        # Precompute M = miller_loop(public_pair)
        circuit = MultiMillerLoopCircuit(
            name="precompute M", curve_id=self.curve_id.value, n_pairs=1
        )
        circuit.write_p_and_q(P=[self.alpha], Q=[self.beta])

        M = E12D("alpha_beta_miller_loop_result", circuit.miller_loop(n_pairs=1))
        gamma_g2 = G2PointCircuit.from_G2Point("gamma_g2", self.gamma)
        delta_g2 = G2PointCircuit.from_G2Point("delta_g2", self.delta)
        ic = StructArray(
            name="ic",
            elmts=[
                G1PointCircuit.from_G1Point(f"ic_{i}", point)
                for i, point in enumerate(self.ic)
            ],
        )
        code = f"""
        pub const vk:Groth16VerifyingKey = Groth16VerifyingKey{{
            alpha_beta_miller_loop_result: {M.serialize(raw=True)},
            gamma_g2: {gamma_g2.serialize(raw=True)},
            delta_g2: {delta_g2.serialize(raw=True)}
        }};

        pub const ic: [G1Point; {len(self.ic)}] = {ic.serialize(raw=True, const=True)};

        """
        return code

    def flatten(self) -> list[int]:
        lst = []
        lst.extend([self.alpha.x, self.alpha.y])
        lst.extend([self.beta.x[0], self.beta.x[1], self.beta.y[0], self.beta.y[1]])
        lst.extend([self.gamma.x[0], self.gamma.x[1], self.gamma.y[0], self.gamma.y[1]])
        lst.extend([self.delta.x[0], self.delta.x[1], self.delta.y[0], self.delta.y[1]])
        for point in self.ic:
            lst.extend([point.x, point.y])
        return lst


def reverse_byte_order_uint256(value: int | bytes) -> int:
    if isinstance(value, int):
        value_bytes = value.to_bytes(32, byteorder="big")
    else:
        value_bytes = value.ljust(
            32, b"\x00"
        )  # Ensure 32 bytes, pad with zeros if needed
    return int.from_bytes(value_bytes[::-1], byteorder="big")


def split_digest(digest: int | bytes):
    reversed_digest = reverse_byte_order_uint256(digest)
    return split_128(reversed_digest)


@dataclasses.dataclass(slots=True)
class Groth16Proof:
    a: G1Point
    b: G2Point
    c: G1Point
    public_inputs: List[int] = dataclasses.field(default_factory=list)
    curve_id: CurveID = None
    image_id: bytes = None  # Only used for risc0 proofs
    journal: bytes = None  # Only used for risc0 proofs
    public_inputs_sp1: bytes = None  # Only used for sp1 proofs

    def __post_init__(self):
        assert (
            self.a.curve_id == self.b.curve_id == self.c.curve_id
        ), f"All points must be on the same curve, got {self.a.curve_id}, {self.b.curve_id}, {self.c.curve_id}"
        self.curve_id = self.a.curve_id
        for pub in self.public_inputs:
            assert (
                0 <= pub < CURVES[self.curve_id.value].n
            ), f"Public input {pub} is out of bounds for curve {self.curve_id}"
        if self.public_inputs_sp1 is not None:
            assert isinstance(
                self.public_inputs_sp1, bytes
            ), "SP1 public inputs must be a bytes object"
            assert (
                len(self.public_inputs_sp1) % 32 == 0
            ), "SP1 public inputs must be a multiple of 32 bytes"

    def from_dict(
        data: dict, public_inputs: None | list | dict = None
    ) -> "Groth16Proof":
        curve_id = try_guessing_curve_id_from_json(data)
        try:
            proof = find_item_from_key_patterns(data, ["proof"])
        except KeyPatternNotFound:
            proof = data

        try:
            seal = find_item_from_key_patterns(data, ["seal"])
            image_id = find_item_from_key_patterns(data, ["image_id"])
            journal = find_item_from_key_patterns(data, ["journal"])

            print("Seal: {}\nImage_id: {}\nJournal: {}".format(seal, image_id, journal))
            return Groth16Proof._from_risc0(
                seal=codecs.decode(seal[2:].replace("\\x", ""), "hex"),
                image_id=codecs.decode(image_id[2:].replace("\\x", ""), "hex"),
                journal=codecs.decode(journal[2:].replace("\\x", ""), "hex"),
            )
        except (ValueError, KeyError, KeyPatternNotFound):
            pass
        except Exception as e:
            print(f"Error when attempmting to parse as Risc0 proof: {e}")
            raise e

        try:
            sp1_vkey = find_item_from_key_patterns(data, ["vkey"])
            public_inputs = find_item_from_key_patterns(data, ["publicValues"])
            sp1_proof = find_item_from_key_patterns(data, ["proof"])
            return Groth16Proof._from_sp1(
                vkey=codecs.decode(sp1_vkey[2:].replace("\\x", ""), "hex"),
                proof=codecs.decode(sp1_proof[2:].replace("\\x", ""), "hex"),
                public_inputs=codecs.decode(
                    public_inputs[2:].replace("\\x", ""), "hex"
                ),
            )
        except KeyPatternNotFound:
            pass
        except Exception as e:
            print(f"Error when attempmting to parse as SP1 proof: {e}")
            raise e

        if public_inputs is not None:
            if isinstance(public_inputs, dict):
                public_inputs = list(public_inputs.values())
            elif isinstance(public_inputs, list):
                pass
            else:
                raise ValueError(f"Invalid public inputs format: {public_inputs}")
        else:
            try:
                public_inputs = find_item_from_key_patterns(data, ["public"])
            except KeyPatternNotFound as e:
                print(f"Error: {e}")
                raise e
        return Groth16Proof(
            a=try_parse_g1_point_from_key(proof, ["a"], curve_id),
            b=try_parse_g2_point_from_key(proof, ["b"], curve_id),
            c=try_parse_g1_point_from_key(proof, ["c", "Krs"], curve_id),
            public_inputs=[io.to_int(pub) for pub in public_inputs],
        )

    def from_json(
        proof_path: str | Path, public_inputs_path: str | Path = None
    ) -> "Groth16Proof":
        path = Path(proof_path)
        try:
            with path.open("r") as f:
                data = json.load(f)
        except FileNotFoundError:
            raise FileNotFoundError(f"The file {proof_path} was not found.")
        except json.JSONDecodeError:
            raise ValueError(f"The file {proof_path} does not contain valid JSON.")
        try:
            if public_inputs_path is not None:
                with Path(public_inputs_path).open("r") as f:
                    public_inputs = json.load(f)
            else:
                public_inputs = None
        except FileNotFoundError:
            raise FileNotFoundError(f"The file {public_inputs_path} was not found.")
        except json.JSONDecodeError:
            raise ValueError(
                f"The file {public_inputs_path} does not contain valid JSON."
            )
        return Groth16Proof.from_dict(data, public_inputs)

    def _from_risc0(
        seal: bytes,
        image_id: bytes,
        journal: bytes,
        CONTROL_ROOT: int = RISC0_CONTROL_ROOT,
        BN254_CONTROL_ID: int = RISC0_BN254_CONTROL_ID,
    ) -> "Groth16Proof":
        assert len(image_id) <= 32, "image_id must be 32 bytes"
        CONTROL_ROOT_0, CONTROL_ROOT_1 = split_digest(CONTROL_ROOT)
        proof = seal[4:]
        journal_digest = hashlib.sha256(journal).digest()
        claim_digest = ok(image_id, journal_digest).digest()
        claim0, claim1 = split_digest(claim_digest)
        return Groth16Proof(
            a=G1Point(
                x=int.from_bytes(proof[0:32], "big"),
                y=int.from_bytes(proof[32:64], "big"),
                curve_id=CurveID.BN254,
            ),
            b=G2Point(
                x=(
                    int.from_bytes(proof[96:128], "big"),
                    int.from_bytes(proof[64:96], "big"),
                ),
                y=(
                    int.from_bytes(proof[160:192], "big"),
                    int.from_bytes(proof[128:160], "big"),
                ),
                curve_id=CurveID.BN254,
            ),
            c=G1Point(
                x=int.from_bytes(proof[192:224], "big"),
                y=int.from_bytes(proof[224:256], "big"),
                curve_id=CurveID.BN254,
            ),
            public_inputs=[
                CONTROL_ROOT_0,
                CONTROL_ROOT_1,
                claim0,
                claim1,
                BN254_CONTROL_ID,
            ],
            image_id=image_id,
            journal=journal,
        )

    def _from_sp1(
        vkey: bytes,
        public_inputs: bytes,
        proof: bytes,
    ) -> "Groth16Proof":
        _selector, proof = proof[:4], proof[4:]
        assert (
            _selector == SP1_VERIFIER_HASH[:4]
        ), f"Invalid SP1 proof version. Expected {SP1_VERIFIER_HASH[:4]} for version {SP1_VERIFIER_VERSION}, got {_selector}\n Please use SP1 verifier version {SP1_VERIFIER_VERSION} or contact garaga developers to update the SP1 verifier version."
        pub_input_hash: bytes = hashlib.sha256(public_inputs).digest()
        pub_input_hash: int = int.from_bytes(pub_input_hash, "big") % (
            2**253
        )  # <=> & ((1 << 253) - 1)
        # print(f"pub_input_hash: {hex(pub_input_hash)}")
        return Groth16Proof(
            a=G1Point(
                x=int.from_bytes(proof[0:32], "big"),
                y=int.from_bytes(proof[32:64], "big"),
                curve_id=CurveID.BN254,
            ),
            b=G2Point(
                x=(
                    int.from_bytes(proof[96:128], "big"),
                    int.from_bytes(proof[64:96], "big"),
                ),
                y=(
                    int.from_bytes(proof[160:192], "big"),
                    int.from_bytes(proof[128:160], "big"),
                ),
                curve_id=CurveID.BN254,
            ),
            c=G1Point(
                x=int.from_bytes(proof[192:224], "big"),
                y=int.from_bytes(proof[224:256], "big"),
                curve_id=CurveID.BN254,
            ),
            public_inputs=[int.from_bytes(vkey, "big"), pub_input_hash],
            public_inputs_sp1=public_inputs,
        )

    def serialize_to_calldata(self) -> list[int]:
        cd = []
        cd.extend(io.bigint_split(self.a.x))
        cd.extend(io.bigint_split(self.a.y))
        cd.extend(io.bigint_split(self.b.x[0]))
        cd.extend(io.bigint_split(self.b.x[1]))
        cd.extend(io.bigint_split(self.b.y[0]))
        cd.extend(io.bigint_split(self.b.y[1]))
        cd.extend(io.bigint_split(self.c.x))
        cd.extend(io.bigint_split(self.c.y))
        if self.image_id and self.journal and self.public_inputs_sp1 is None:
            # Risc0 mode.
            # Public inputs will be reconstructed from image id and journal.
            image_id_u256 = io.bigint_split(
                int.from_bytes(self.image_id, "big"), 8, 2**32
            )[::-1]
            journal = list(self.journal)
            # Span of u32, length 8.
            cd.append(8)
            cd.extend(image_id_u256)
            # Span of u8, length depends on input
            cd.append(len(self.journal))
            cd.extend(journal)
        elif self.public_inputs_sp1 is not None:
            # SP1 mode.
            # Public inputs 0 is vkey
            cd.extend(io.bigint_split(self.public_inputs[0], 2, 2**128))
            # Public inputs 1 is sha256(public_inputs_sp1) % 2**253
            # We serialize pub_inputs_sp1 as u32 array for Cairo sha256.
            n_bytes = len(self.public_inputs_sp1)
            assert n_bytes % 32 == 0, "SP1 public inputs must be a multiple of 32 bytes"
            n_words = n_bytes // 32
            cd.append(n_words)  # Deserialization 8 by 8
            for i in range(0, n_bytes, 4):
                cd.append(int.from_bytes(self.public_inputs_sp1[i : i + 4], "big"))
        else:
            # Normal mode.
            cd.append(len(self.public_inputs))
            for pub in self.public_inputs:
                cd.extend(io.bigint_split(pub, 2, 2**128))
        return cd

    def flatten(self, include_public_inputs: bool = True) -> list[int]:
        lst = []
        lst.extend([self.a.x, self.a.y])
        lst.extend([self.b.x[0], self.b.x[1]])
        lst.extend([self.b.y[0], self.b.y[1]])
        lst.extend([self.c.x, self.c.y])
        if include_public_inputs:
            lst.extend(self.public_inputs)
        return lst


class ExitCode:
    def __init__(self, system, user):
        self.system = system
        self.user = user


class Output:
    def __init__(self, journal_digest, assumptions_digest):
        self.journal_digest = journal_digest
        self.assumptions_digest = assumptions_digest

    def digest(self):
        return hashlib.sha256(
            hashlib.sha256(b"risc0.Output").digest()
            + self.journal_digest
            + self.assumptions_digest
            + (2 << 8).to_bytes(2, byteorder="big")
        ).digest()


class ReceiptClaim:
    def __init__(
        self,
        pre_state_digest,
        post_state_digest,
        exit_code,
        input,
        output,
        tag_digest: bytes = hashlib.sha256(b"risc0.ReceiptClaim").digest(),
    ):
        self.pre_state_digest = pre_state_digest
        self.post_state_digest = post_state_digest
        self.exit_code = exit_code
        self.input = input
        self.output = output.digest()
        self.TAG_DIGEST = tag_digest

    def digest(self):
        # print(f"TAG_DIGEST: {self.TAG_DIGEST.hex()}")
        # print(f"self.input: {self.input.hex()}")
        # print(f"self.pre_state_digest: {self.pre_state_digest.hex()}")
        # print(f"self.post_state_digest: {self.post_state_digest.hex()}")
        # print(f"self.output: {self.output.hex()}")
        # print(f"self.exit_code.system: {self.exit_code.system}")
        # print(f"self.exit_code.user: {self.exit_code.user}")
        return hashlib.sha256(
            self.TAG_DIGEST
            + self.input
            + self.pre_state_digest
            + self.post_state_digest
            + self.output
            + (self.exit_code.system << 24).to_bytes(4, byteorder="big")
            + (self.exit_code.user << 24).to_bytes(4, byteorder="big")
            + (4 << 8).to_bytes(2, byteorder="big")
        ).digest()


def ok(image_id, journal_digest):
    SYSTEM_STATE_ZERO_DIGEST = bytes.fromhex(
        "A3ACC27117418996340B84E5A90F3EF4C49D22C79E44AAD822EC9C313E1EB8E2"
    )  # https://github.com/risc0/risc0-ethereum/blob/34d2fee4ca6b5fb354a8a1a00c43f8945097bfe5/contracts/src/IRiscZeroVerifier.sol#L60
    return ReceiptClaim(
        pre_state_digest=image_id,
        post_state_digest=SYSTEM_STATE_ZERO_DIGEST,
        exit_code=ExitCode(0, 0),  # (Halted, 0)
        input=bytes(32),  # bytes32(0)
        output=Output(journal_digest, bytes(32)),  # Output(journalDigest, bytes32(0))
    )


if __name__ == "__main__":
    # PATH = Path(__file__).parent
    # print(f"PATH: {PATH}")
    # proof = Groth16Proof.from_json(
    #     f"{PATH}/examples/gnark_proof_bn254.json",
    #     f"{PATH}/examples/gnark_public_bn254.json",
    # )
    # # print(proof)
    # vk = Groth16VerifyingKey.from_json(f"{PATH}/examples/gnark_vk_bn254.json")
    # print(vk)

    # vk_risc0 = Groth16VerifyingKey.from_json(f"{PATH}/examples/vk_risc0.json")
    # # print(vk_risc0)

    # Risc0 proof extracted from https://sepolia.etherscan.io/tx/0x2308aeefef309097aeaf0e3660915d5b80813e693ac72147b651e0196155235d
    bproof = bytes.fromhex(
        "310fe5982466f8f1bab4d00a829cafcda46036fb9c5108df341746ab5f7532aa71aee03b0947eaf1af095584de8d5bd0a91a811f071a555c21a113476aa167108dfeb73913c3a1ef6a5baac68cddd25fafdbf660c4e479f7a836cc1b98904610ead5c9ab2c62f6fdf8ca099964080c95beebf5728b41728128ec0c7823f8adf22e5bfeed1110de7c21ed2dc1e2fd8f2c52d68a15129cf68f18a3087131920e8dcb40a81b003f524b6dcbabdc1e270494bc39b190bddfdb13106409350f80b6204d89da4c16842b4139dd02a39829cf1403657ad00080300a32148c31093cb752809cae2e075db0a79893a6d71a4a7d61111fdff741aeb198dd7fb00b4fa23714ddfd8093"
    )

    def parse_proof_and_signals(proof: bytes) -> Groth16Proof:
        proof = proof[4:]

        def bytes_32_iterator(data: bytes):
            for i in range(0, len(data), 32):
                yield io.to_int(data[i : i + 32])

        it = bytes_32_iterator(proof)
        print(len(proof) / 32)
        _a = G1Point(x=next(it), y=next(it), curve_id=CurveID.BN254)
        _b = G2Point(
            x=[next(it), next(it)][::-1],
            y=[next(it), next(it)][::-1],
            curve_id=CurveID.BN254,
        )
        _c = G1Point(x=next(it), y=next(it), curve_id=CurveID.BN254)

        public_inputs = list(it)
        print(len(public_inputs))
        return Groth16Proof(
            a=_a,
            b=_b,
            c=_c,
            public_inputs=list(it),
        )

    proof = parse_proof_and_signals(bproof)
    print(proof)



import binascii
import functools
import hashlib
import random
from dataclasses import dataclass
from enum import Enum
from typing import List, Optional, Union

import requests
from requests.exceptions import RequestException

from garaga.definitions import CurveID, Fp2, G1Point, G2Point, get_base_field
from garaga.hints import io


class DrandNetwork(Enum):
    default = "8990e7a9aaed2ffed73dbd7092123d6f289930540d7651336225dc172e51b2ce"
    quicknet = "52db9ba70e0cc0f6eaf7803dd07447a1f5477735fd3f661792ba94600c84e971"


def digest_func(round_number: int) -> bytes:
    assert isinstance(round_number, int)
    assert 0 <= round_number < 2**64
    h = hashlib.sha256()
    h.update(round_number.to_bytes(8, byteorder="big"))
    return h.digest()


@dataclass
class NetworkInfo:
    public_key: Union[G1Point, G2Point]
    period: int
    genesis_time: int
    hash: str
    group_hash: str
    scheme_id: str
    beacon_id: Optional[str] = None


@dataclass
class RandomnessBeacon:
    round_number: int
    randomness: int
    signature: str
    previous_signature: Optional[str] = None

    @property
    def signature_point(self) -> G1Point | G2Point:
        return deserialize_bls_point(bytes.fromhex(self.signature))


BASE_URLS = [
    "https://drand.cloudflare.com",
    "https://api.drand.sh",
    "https://api2.drand.sh",
    "https://api3.drand.sh",
]


def make_request(endpoint: str) -> requests.Response:
    # Create a copy and shuffle instead of random.choice for better distribution
    base_urls = BASE_URLS.copy()
    random.shuffle(base_urls)

    last_error = None
    for url in base_urls:
        try:
            response = requests.get(f"{url}{endpoint}", timeout=5)  # Add timeout
            response.raise_for_status()
            return response
        except Exception as e:
            print(f"Request to {url} failed: {str(e)}")
            last_error = e
            continue

    raise RequestException(f"All URLs failed. Last error: {last_error}")


def deserialize_bls_point(s_string: bytes) -> Union[G1Point, G2Point]:
    m_byte = s_string[0] & 0xE0
    if m_byte in (0x20, 0x60, 0xE0):
        raise ValueError("Invalid encoding")

    C_bit = (m_byte & 0x80) >> 7  # Compression bit
    I_bit = (m_byte & 0x40) >> 6  # Infinity bit
    S_bit = (m_byte & 0x20) >> 5  # Sign bit

    s_string = bytes([s_string[0] & 0x1F]) + s_string[1:]

    if I_bit == 1:
        if any(b != 0 for b in s_string):
            raise ValueError("Invalid encoding for point at infinity")
        return (
            G1Point.infinity(CurveID.BLS12_381)
            if len(s_string) in (48, 96)
            else G2Point.infinity(CurveID.BLS12_381)
        )

    if C_bit == 0:
        if len(s_string) == 96:  # G1 point (uncompressed)
            x = int.from_bytes(s_string[:48], "big")
            y = int.from_bytes(s_string[48:], "big")
            return G1Point(x, y, CurveID.BLS12_381)
        elif len(s_string) == 192:  # G2 point (uncompressed)
            x0 = int.from_bytes(s_string[:48], "big")
            x1 = int.from_bytes(s_string[48:96], "big")
            y0 = int.from_bytes(s_string[96:144], "big")
            y1 = int.from_bytes(s_string[144:], "big")
            return G2Point((x0, x1), (y0, y1), CurveID.BLS12_381)
        else:
            raise ValueError(f"Invalid length for uncompressed point: {len(s_string)}")
    else:  # C_bit == 1
        x = int.from_bytes(s_string, "big")
        if len(s_string) == 48:  # G1 point (compressed)
            field = get_base_field(CurveID.BLS12_381)
            y2 = field(x**3 + 4)
            if S_bit == 1:
                y = y2.sqrt(min_root=False)
            else:
                y = y2.sqrt(min_root=True)
            return G1Point(x, y.value, CurveID.BLS12_381)
        elif len(s_string) == 96:  # G2 point (compressed)
            field = get_base_field(CurveID.BLS12_381, Fp2)
            x = field((x & ((1 << 384) - 1), x >> 384))
            y2 = x**3 + field((4, 4))
            y = y2.sqrt()

            if S_bit == 1:
                # Choose largest root
                # print("choose largest")
                if not y.lexicographically_largest():
                    # print("root was not largest, negating")
                    y = -y
            else:
                # Choose smallest root
                # print("choose smallest")
                if y.lexicographically_largest():
                    # print("root was largest, negating")
                    y = -y

            return G2Point(
                (x.a0.value, x.a1.value),
                (y.a0.value, y.a1.value),
                CurveID.BLS12_381,
            )
        else:
            raise ValueError(f"Invalid length for compressed point: {len(s_string)}")


def get_chains() -> List[str]:
    response = make_request("/chains")
    all_chains = response.json()
    # only keep chains present in DrandNetwork enum
    chains = [
        chain
        for chain in all_chains
        if chain in [network.value for network in DrandNetwork]
    ]
    return chains


def get_chain_info(chain_hash: str) -> NetworkInfo:
    response = make_request(f"/{chain_hash}/info")
    data = response.json()

    # Parse the public key
    public_key_hex = data["public_key"]
    public_key_bytes = binascii.unhexlify(public_key_hex)

    try:
        public_key = deserialize_bls_point(public_key_bytes)
    except Exception as e:
        public_key = G2Point.infinity(CurveID.BLS12_381)

    return NetworkInfo(
        public_key=public_key,
        period=data["period"],
        genesis_time=data["genesis_time"],
        hash=data["hash"],
        group_hash=data["groupHash"],
        scheme_id=data["schemeID"],
        beacon_id=data.get("metadata", {}).get("beaconID"),
    )


def get_latest_randomness(chain_hash: str) -> RandomnessBeacon:
    response = make_request(f"/{chain_hash}/public/latest")
    return _parse_randomness_beacon(response.json())


def get_randomness(chain_hash: str, round_number: int) -> RandomnessBeacon:
    response = make_request(f"/{chain_hash}/public/{round_number}")
    return _parse_randomness_beacon(response.json())


def _parse_randomness_beacon(data: dict) -> RandomnessBeacon:
    return RandomnessBeacon(
        round_number=io.to_int(data["round"]),
        randomness=int(data["randomness"], 16),
        signature=data["signature"],
        previous_signature=data.get("previous_signature"),
    )


@functools.lru_cache(maxsize=None)
def print_all_chain_info() -> dict[DrandNetwork, NetworkInfo]:
    # Try multiple times to get chains
    max_retries = 3
    for attempt in range(max_retries):
        try:
            chains = get_chains()
            if not chains:
                raise ValueError("No chains returned")

            print(f"Found {len(chains)} chains: {chains}")
            print("-" * 40)

            chain_infos = {}
            expected_networks = set(DrandNetwork)

            for chain_hash in chains:
                try:
                    info = get_chain_info(chain_hash)
                    try:
                        network = DrandNetwork(chain_hash)
                        chain_infos[network] = info
                        expected_networks.remove(network)

                        print(f"Chain: {chain_hash}")
                        print(f"  Public Key: {info.public_key}")
                        print(f"  Period: {info.period} seconds")
                        print(f"  Genesis Time: {info.genesis_time}")
                        print(f"  Hash: {info.hash}")
                        print(f"  Group Hash: {info.group_hash}")
                        print(f"  Scheme ID: {info.scheme_id}")
                        if info.beacon_id:
                            print(f"  Beacon ID: {info.beacon_id}")
                        print("-" * 40)
                    except ValueError:
                        print(f"Unknown chain hash: {chain_hash}")
                        continue
                except Exception as e:
                    print(f"Error fetching info for chain {chain_hash}: {str(e)}")
                    continue

            # If we found all expected networks, return the results
            if not expected_networks:
                return chain_infos

            print(f"Missing networks: {[n.name for n in expected_networks]}")

        except Exception as e:
            print(f"Attempt {attempt + 1}/{max_retries} failed: {str(e)}")
            if attempt < max_retries - 1:
                print("Retrying...")
                continue

    raise RuntimeError(
        f"Failed to get complete chain info after {max_retries} attempts"
    )


# Example usage
if __name__ == "__main__":
    chain_infos = print_all_chain_info()
    from garaga.modulo_circuit_structs import StructArray
    from garaga.signature import hash_to_curve

    # latest_randomness = get_latest_randomness(chain_infos[DrandNetwork.default].hash)
    # print("Latest randomness:", latest_randomness)
    network = DrandNetwork.quicknet
    chain = chain_infos[network]
    print(chain.public_key)
    print(-chain.public_key)

    round = get_randomness(chain.hash, 1000)
    print("Randomness for round 1000:", round)
    sha256 = hashlib.sha256()
    sha256.update(bytes.fromhex(round.signature))
    print("randomness", sha256.hexdigest())
    print("random beacon", hex(round.randomness))

    msg_point = hash_to_curve(
        digest_func(round.round_number), CurveID.BLS12_381, "sha256"
    )
    print("message", msg_point)

    from garaga.definitions import G1G2Pair
    from garaga.modulo_circuit_structs import G2Line, StructArray
    from garaga.precompiled_circuits.multi_miller_loop import precompute_lines

    print(
        "pairing",
        G1G2Pair.pair(
            [
                G1G2Pair(
                    p=round.signature_point, q=G2Point.get_nG(CurveID.BLS12_381, 1)
                ),
                G1G2Pair(p=msg_point, q=-chain.public_key),
            ],
            curve_id=CurveID.BLS12_381,
        ).value_coeffs,
    )

    lines = precompute_lines([G2Point.get_nG(CurveID.BLS12_381, 1), -chain.public_key])
    precomputed_lines = StructArray(
        name="lines",
        elmts=[
            G2Line(name=f"line{i}", elmts=lines[i : i + 4])
            for i in range(0, len(lines), 4)
        ],
    )

    def generate_precomputed_lines_code(precomputed_lines: StructArray) -> str:
        return f"pub const precomputed_lines: [G2Line; {len(precomputed_lines)//4}] = {precomputed_lines.serialize(raw=True, const=True)};"

    # print(generate_precomputed_lines_code(precomputed_lines))



import functools
from dataclasses import dataclass
from functools import lru_cache

from garaga import garaga_rs
from garaga import modulo_circuit_structs as structs
from garaga.algebra import Polynomial, PyFelt
from garaga.definitions import CurveID, G1G2Pair, get_base_field, get_irreducible_poly
from garaga.poseidon_transcript import CairoPoseidonTranscript
from garaga.precompiled_circuits.multi_miller_loop import precompute_lines
from garaga.precompiled_circuits.multi_pairing_check import (
    MultiMillerLoopCircuit,
    MultiPairingCheckCircuit,
    WriteOps,
    get_max_Q_degree,
)


@dataclass(slots=True)
class MPCheckCalldataBuilder:
    curve_id: CurveID
    pairs: list[G1G2Pair]
    n_fixed_g2: int
    public_pair: G1G2Pair | None

    def __hash__(self):
        return hash(
            (self.curve_id, tuple(self.pairs), self.n_fixed_g2, self.public_pair)
        )

    def __post_init__(self):
        # Validate input
        assert isinstance(self.pairs, (list, tuple))
        assert all(
            isinstance(pair, G1G2Pair) for pair in self.pairs
        ), f"All pairs must be G1G2Pair, got {[type(pair) for pair in self.pairs]}"
        assert all(
            self.curve_id == pair.curve_id == self.pairs[0].curve_id
            for pair in self.pairs
        ), f"All pairs must be on the same curve, got {[pair.curve_id for pair in self.pairs]}"
        assert (
            isinstance(self.public_pair, G1G2Pair) or self.public_pair is None
        ), f"Extra pair must be G1G2Pair or None, got {self.public_pair}"
        assert len(self.pairs) >= 2
        assert 0 <= self.n_fixed_g2 <= len(self.pairs)

    @property
    def include_miller_loop_result(self):
        return self.public_pair is not None

    @property
    def field(self):
        return get_base_field(self.curve_id)

    @property
    def big_Q_expected_len(self):
        return get_max_Q_degree(self.curve_id.value, len(self.pairs)) + 1

    @lru_cache(maxsize=1)
    def extra_miller_loop_result(self) -> list[PyFelt] | None:
        if self.include_miller_loop_result:
            circuit = MultiMillerLoopCircuit(
                name="precompute M", curve_id=self.curve_id.value, n_pairs=1
            )
            circuit.write_p_and_q_raw(self.public_pair.to_pyfelt_list())
            M = circuit.miller_loop(n_pairs=1)
            return [mi.felt for mi in M]
        else:
            return None

    @lru_cache(maxsize=1)
    def lines(self) -> list[PyFelt]:
        lines = precompute_lines([pair.q for pair in self.pairs[0 : self.n_fixed_g2]])
        assert len(lines) % 4 == 0, f"Lines must be a multiple of 4, got {len(lines)}"
        return lines

    def _init_circuit(self) -> MultiPairingCheckCircuit:
        mpcheck_circuit = MultiPairingCheckCircuit(
            name="mpcheck",
            curve_id=self.curve_id.value,
            n_pairs=len(self.pairs),
            precompute_lines=bool(self.n_fixed_g2),
            n_points_precomputed_lines=self.n_fixed_g2,
        )

        mpcheck_circuit.precomputed_lines = mpcheck_circuit.write_elements(
            self.lines(), WriteOps.INPUT
        )
        mpcheck_circuit._precomputed_lines_generator = (
            mpcheck_circuit._create_precomputed_lines_generator()
        )

        p_q_input = []
        for pair in self.pairs:
            p_q_input.extend(pair.to_pyfelt_list())
        mpcheck_circuit.write_p_and_q_raw(p_q_input)
        return mpcheck_circuit

    def _retrieve_Pis_Qis_and_Ris_from_circuit(
        self, mpcheck_circuit: MultiPairingCheckCircuit
    ) -> tuple[list[list[PyFelt]], list[list[PyFelt]], list[list[PyFelt]]]:
        relations = mpcheck_circuit.accumulate_poly_instructions[0]
        Qis = relations.Qis
        Pis = relations.Pis
        Ris = relations.Ris
        return Pis, Qis, Ris

    def _get_passed_Ris_from_Ris(self, Ris: list[list[PyFelt]]) -> list[list[PyFelt]]:
        passed_Ris = (
            Ris if self.curve_id == CurveID.BLS12_381 else Ris[1:]
        )  # Skip first Ri for BN254 as it known to be one (lambda_root*lambda_root_inverse) result

        if self.public_pair is not None:
            passed_Ris = passed_Ris[
                :-1
            ]  # Skip last Ri as it is known to be 1 and we use FP12Mul_AssertOne circuit

        return passed_Ris

    def _init_transcript(self) -> CairoPoseidonTranscript:
        init_hash = (
            f"MPCHECK_{self.curve_id.name}_{len(self.pairs)}P_{self.n_fixed_g2}F"
        )
        transcript = CairoPoseidonTranscript(
            init_hash=int.from_bytes(init_hash.encode(), byteorder="big")
        )
        # Hash inputs.
        for pair in self.pairs:
            transcript.hash_limbs_multi(pair.to_pyfelt_list())
        return transcript

    def _hash_hints_and_get_base_random_rlc_coeff(
        self,
        transcript: CairoPoseidonTranscript,
        lambda_root: list[PyFelt],
        lambda_root_inverse: list[PyFelt],
        scaling_factor: list[PyFelt],
        scaling_factor_sparsity: list[PyFelt],
        Ris: list[list[PyFelt]],
    ):
        if self.curve_id == CurveID.BN254:
            transcript.hash_limbs_multi(lambda_root)

        transcript.hash_limbs_multi(lambda_root_inverse)
        transcript.hash_limbs_multi(scaling_factor, sparsity=scaling_factor_sparsity)

        for Ri in Ris:
            assert len(Ri) == 12
            transcript.hash_limbs_multi(Ri)

        return self.field(transcript.s1)

    def _hash_big_Q_and_get_z(
        self, transcript: CairoPoseidonTranscript, big_Q: list[PyFelt]
    ):
        transcript.hash_limbs_multi(big_Q)
        return self.field(transcript.s0)

    def _sanity_check_verify_rlc_equation(
        self,
        z: PyFelt,
        cis: list[PyFelt],
        Pis: list[list[PyFelt]],
        big_Q: Polynomial,
        Ris: list[list[PyFelt]],
    ):
        lhs = self.field.zero()
        for i, ci in enumerate(cis):
            Prod_Pis_of_z = functools.reduce(
                lambda x, y: x * y, [Polynomial(pi).evaluate(z) for pi in Pis[i]]
            )
            Ri_of_z = Polynomial(Ris[i]).evaluate(z)
            lhs += ci * (Prod_Pis_of_z - Ri_of_z)
            # print(f"lhs_{i} : {io.int_to_u384(lhs)}")

        P_irr = get_irreducible_poly(curve_id=self.curve_id, extension_degree=12)
        big_Q_of_z = big_Q.evaluate(z)
        P_of_z = P_irr.evaluate(z)
        # print(f"big_Q_of_z : {io.int_to_u384(big_Q_of_z)}")
        # print(f"P_of_z : {io.int_to_u384(P_of_z)}")
        assert lhs == big_Q_of_z * P_of_z, "Check failed."

    @lru_cache(maxsize=1)
    def build_mpcheck_hint(
        self,
    ) -> tuple[
        structs.Cairo1SerializableStruct, structs.Cairo1SerializableStruct | None
    ]:
        """
        Return MPCheckHint struct and small_Q struct if extra_miller_loop_result is True
        """
        mpcheck_circuit = self._init_circuit()
        transcript = self._init_transcript()

        _, lambda_root, lambda_root_inverse, scaling_factor, scaling_factor_sparsity = (
            mpcheck_circuit.multi_pairing_check(
                len(self.pairs), self.extra_miller_loop_result()
            )
        )
        Pis, Qis, Ris = self._retrieve_Pis_Qis_and_Ris_from_circuit(mpcheck_circuit)
        passed_Ris = self._get_passed_Ris_from_Ris(Ris)

        c0 = self._hash_hints_and_get_base_random_rlc_coeff(
            transcript,
            lambda_root,
            lambda_root_inverse,
            scaling_factor,
            scaling_factor_sparsity,
            passed_Ris,
        )

        n_relations_with_ci = len(passed_Ris) + (
            1 if self.curve_id == CurveID.BN254 else 0
        )

        ci, cis, big_Q = c0, [], Polynomial.zero(self.field.p)
        for i in range(n_relations_with_ci):
            # print(f"c_{i} : {io.int_to_u384(ci)}")
            cis.append(ci)
            big_Q += Qis[i] * ci
            ci *= ci

        big_Q_coeffs = big_Q.get_coeffs()
        big_Q_coeffs.extend(
            [self.field.zero()] * (self.big_Q_expected_len - len(big_Q_coeffs))
        )

        z = self._hash_big_Q_and_get_z(transcript, big_Q_coeffs)
        self._sanity_check_verify_rlc_equation(z, cis, Pis, big_Q, Ris)

        if self.public_pair is None:
            assert [x.value for x in passed_Ris[-1]] == [
                1,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
                0,
            ], f"Last Ri must be 1_Fp12, got {[x.value for x in passed_Ris[-1]]}"
            passed_Ris = passed_Ris[:-1]  # Skip last Ri as it is known to be 1

        if self.curve_id == CurveID.BN254:
            hint_struct_list_init = [
                structs.E12D(name="lambda_root", elmts=lambda_root)
            ]
        else:
            hint_struct_list_init = []

        if self.include_miller_loop_result:
            small_Q = Qis[-1].get_coeffs()
            small_Q = small_Q + [self.field.zero()] * (11 - len(small_Q))
            small_Q_struct = structs.E12DMulQuotient(name="small_Q", elmts=small_Q)
        else:
            small_Q_struct = None

        return (
            structs.Struct(
                struct_name=f"MPCheckHint{self.curve_id.name}",
                name="hint",
                elmts=hint_struct_list_init
                + [
                    structs.E12D(name="lambda_root_inverse", elmts=lambda_root_inverse),
                    structs.MillerLoopResultScalingFactor(
                        name="w",
                        elmts=[
                            wi
                            for wi, si in zip(scaling_factor, scaling_factor_sparsity)
                            if si != 0
                        ],
                    ),
                    structs.StructSpan(
                        name="Ris",
                        elmts=[
                            structs.E12D(name=f"R{i}", elmts=[ri.felt for ri in Ri])
                            for i, Ri in enumerate(passed_Ris)
                        ],
                    ),
                    structs.u384Array(name="big_Q", elmts=big_Q_coeffs),
                ]
                + [structs.felt252(name="z", elmts=[z])],
            ),
            small_Q_struct,
        )

    def _get_input_structs(self) -> list[structs.Cairo1SerializableStruct]:
        inputs = []
        for i, pair in enumerate(self.pairs):
            inputs.append(
                structs.G1G2PairCircuit(name=f"pair{i}", elmts=pair.to_pyfelt_list())
            )
        if self.include_miller_loop_result:
            inputs.append(
                structs.E12D(
                    name="precomputed_miller_loop_result",
                    elmts=self.extra_miller_loop_result(),
                )
            )
        if self.n_fixed_g2 > 0:
            lines = self.lines()
            inputs.append(
                structs.StructSpan(
                    name="lines",
                    elmts=[
                        structs.G2Line(name=f"line{i}", elmts=lines[i : i + 4])
                        for i in range(0, len(lines), 4)
                    ],
                )
            )
        mpcheck_hint, small_Q = self.build_mpcheck_hint()
        inputs.append(mpcheck_hint)
        if small_Q is not None:
            inputs.append(small_Q)
        return inputs

    def to_cairo_1_test(self):
        # print(
        #     f"Generating MPCheck test for {self.curve_id.name} with {len(self.pairs)} pairs and {self.n_fixed_g2} fixed G2 points, using extra Miller loop result: {self.include_miller_loop_result}"
        # )
        with_extra = (
            "_with_extra_miller_loop_result" if self.include_miller_loop_result else ""
        )
        fn_name = f"multi_pairing_check_{self.curve_id.name.lower()}_{len(self.pairs)}P_{self.n_fixed_g2}F{with_extra}"
        inputs = self._get_input_structs()

        test_name = f"test_{self.curve_id.name}_mpcheck_{len(self.pairs)}P_{self.n_fixed_g2}F{with_extra}"

        input_code = ""
        for struct in inputs:
            input_code += struct.serialize()

        code = f"""
        #[test]
        fn {test_name}() {{
            {input_code}
            let res = {fn_name}({', '.join([struct.name for struct in inputs])});
            assert!(res);
        }}
        """
        return code

    def _serialize_to_calldata_rust(self) -> list[int]:
        return garaga_rs.mpc_calldata_builder(
            self.curve_id.value,
            [element.value for pair in self.pairs for element in pair.to_pyfelt_list()],
            self.n_fixed_g2,
            (
                [element.value for element in self.public_pair.to_pyfelt_list()]
                if self.public_pair is not None
                else []
            ),
        )

    def serialize_to_calldata(
        self,
        use_rust=False,
    ) -> list[int]:
        if use_rust:
            return self._serialize_to_calldata_rust()

        mpcheck_hint, small_Q = self.build_mpcheck_hint()

        call_data: list[int] = []
        call_data.extend(mpcheck_hint.serialize_to_calldata())
        if small_Q is not None:
            call_data.extend(small_Q.serialize_to_calldata())

        return call_data



from typing import Union

from garaga import garaga_rs
from garaga.definitions import G1G2Pair, ProofSystem
from garaga.precompiled_circuits.honk import (
    CONST_PROOF_SIZE_LOG_N,
    G2_POINT_KZG_1,
    G2_POINT_KZG_2,
    NUMBER_OF_ENTITIES,
    CurveID,
    G1Point,
    HonkProof,
    HonkTranscript,
    HonkVerifierCircuits,
    HonkVk,
    ModuloCircuitElement,
    ZKHonkProof,
    ZKHonkTranscript,
    ZKHonkVerifierCircuits,
)
from garaga.starknet.tests_and_calldata_generators.mpcheck import MPCheckCalldataBuilder
from garaga.starknet.tests_and_calldata_generators.msm import MSMCalldataBuilder


def filter_msm_scalars(
    scalars: list[ModuloCircuitElement], log_n: int, zk: bool = False
) -> list[ModuloCircuitElement]:
    if zk:
        assert len(scalars) == NUMBER_OF_ENTITIES + CONST_PROOF_SIZE_LOG_N + 3 + 3
        start_dummy = 1 + NUMBER_OF_ENTITIES + log_n
        end_dummy = 1 + NUMBER_OF_ENTITIES + CONST_PROOF_SIZE_LOG_N
    else:
        assert len(scalars) == NUMBER_OF_ENTITIES + CONST_PROOF_SIZE_LOG_N + 2
        start_dummy = NUMBER_OF_ENTITIES + log_n
        end_dummy = NUMBER_OF_ENTITIES + CONST_PROOF_SIZE_LOG_N

    # Verify zeros in dummy section
    assert all(
        s.value == 0 for s in scalars[start_dummy:end_dummy]
    ), "Expected all dummy round scalars to be 0"

    # Keep everything except dummy section
    scalars_no_dummy = scalars[:start_dummy] + scalars[end_dummy:]
    # Remove the first element (== 1)
    scalars_filtered = scalars_no_dummy[1:]
    scalars_filtered_no_nones = [
        scalar for scalar in scalars_filtered if scalar is not None
    ]
    return scalars_filtered_no_nones


def extract_msm_scalars(
    scalars: list[ModuloCircuitElement], log_n: int, zk: bool = False
) -> list[ModuloCircuitElement]:
    scalars_msm = filter_msm_scalars(scalars, log_n, zk)
    return [s.value for s in scalars_msm]


def get_ultra_flavor_honk_calldata_from_vk_and_proof(
    vk: HonkVk,
    proof: Union[HonkProof, ZKHonkProof],
    system: ProofSystem = ProofSystem.UltraKeccakHonk,
    use_rust: bool = False,
) -> list[int]:
    if use_rust:
        return _honk_calldata_from_vk_and_proof_rust(vk, proof, system)

    zk = system in [ProofSystem.UltraKeccakZKHonk, ProofSystem.UltraStarknetZKHonk]

    if zk:
        assert isinstance(proof, ZKHonkProof)
        tp = ZKHonkTranscript.from_proof(vk, proof, system)
        circuit = ZKHonkVerifierCircuits(name="test", log_n=vk.log_circuit_size)
    else:
        assert isinstance(proof, HonkProof)
        tp = HonkTranscript.from_proof(vk, proof, system)
        circuit = HonkVerifierCircuits(name="test", log_n=vk.log_circuit_size)

    vk_circuit = vk.to_circuit_elements(circuit)
    proof_circuit = proof.to_circuit_elements(circuit)
    tp = tp.to_circuit_elements(circuit)

    if zk:
        vanishing_check, diff_check = circuit.check_evals_consistency(
            proof_circuit.libra_evaluation,
            proof_circuit.libra_poly_evals,
            tp.gemini_r,
            tp.sum_check_u_challenges,
        )
        assert vanishing_check.value != 0
        assert diff_check.value == 0

    if zk:
        scalars = circuit.compute_shplemini_msm_scalars(
            proof_circuit.sumcheck_evaluations,
            proof_circuit.gemini_masking_eval,
            proof_circuit.gemini_a_evaluations,
            proof_circuit.libra_poly_evals,
            tp.gemini_r,
            tp.rho,
            tp.shplonk_z,
            tp.shplonk_nu,
            tp.sum_check_u_challenges,
        )
    else:
        scalars = circuit.compute_shplemini_msm_scalars(
            proof_circuit.sumcheck_evaluations,
            proof_circuit.gemini_a_evaluations,
            tp.gemini_r,
            tp.rho,
            tp.shplonk_z,
            tp.shplonk_nu,
            tp.sum_check_u_challenges,
        )

    scalars_msm = extract_msm_scalars(scalars, vk.log_circuit_size, zk)
    # Swap last two scalars :
    scalars_msm = scalars_msm[:-2] + scalars_msm[-2:][::-1]

    if zk:
        # Place first scalar just after the vk_lagrange_last point (index 27)
        first_scalar = scalars_msm.pop(0)
        scalars_msm.insert(27, first_scalar)

    if zk:
        points = [
            vk.qm,  # 0
            vk.qc,  # 1
            vk.ql,  # 2
            vk.qr,  # 3
            vk.qo,  # 4
            vk.q4,  # 5
            vk.qLookup,  # 6
            vk.qArith,  # 7
            vk.qDeltaRange,  # 8
            vk.qElliptic,  # 9
            vk.qAux,  # 10
            vk.qPoseidon2External,  # 11
            vk.qPoseidon2Internal,  # 12
            vk.s1,  # 13
            vk.s2,  # 14
            vk.s3,  # 15
            vk.s4,  # 16
            vk.id1,  # 17
            vk.id2,  # 18
            vk.id3,  # 19
            vk.id4,  # 20
            vk.t1,  # 21
            vk.t2,  # 22
            vk.t3,  # 23
            vk.t4,  # 24
            vk.lagrange_first,  # 25
            vk.lagrange_last,  # 26
            proof.gemini_masking_poly,  # 27
            proof.w1,  # 28
            proof.w2,  # 29
            proof.w3,  # 30
            proof.w4,  # 31
            proof.z_perm,  # 32
            proof.lookup_inverses,  # 33
            proof.lookup_read_counts,  # 34
            proof.lookup_read_tags,  # 35
        ]
        points.extend(proof.gemini_fold_comms[: vk.log_circuit_size - 1])
        points.extend(proof.libra_commitments)
        points.append(proof.kzg_quotient)
        points.append(G1Point.get_nG(CurveID.BN254, 1))

    else:
        points = [
            vk.qm,  # 1
            vk.qc,  # 2
            vk.ql,  # 3
            vk.qr,  # 4
            vk.qo,  # 5
            vk.q4,  # 6
            vk.qLookup,  # 7
            vk.qArith,  # 8
            vk.qDeltaRange,  # 9
            vk.qElliptic,  # 10
            vk.qAux,  # 11
            vk.qPoseidon2External,  # 12
            vk.qPoseidon2Internal,  # 13
            vk.s1,  # 14
            vk.s2,  # 15
            vk.s3,  # 16
            vk.s4,  # 17
            vk.id1,  # 18
            vk.id2,  # 19
            vk.id3,  # 20
            vk.id4,  # 21
            vk.t1,  # 22
            vk.t2,  # 23
            vk.t3,  # 24
            vk.t4,  # 25
            vk.lagrange_first,  # 26
            vk.lagrange_last,  # 27
            proof.w1,  # 28
            proof.w2,  # 29
            proof.w3,  # 30
            proof.w4,  # 31
            proof.z_perm,  # 32
            proof.lookup_inverses,  # 33
            proof.lookup_read_counts,  # 34
            proof.lookup_read_tags,  # 35
        ]
        points.extend(proof.gemini_fold_comms[: vk.log_circuit_size - 1])
        points.append(proof.kzg_quotient)
        points.append(G1Point.get_nG(CurveID.BN254, 1))

    msm_builder = MSMCalldataBuilder(
        CurveID.BN254,
        points=points,
        scalars=scalars_msm,
    )

    P_0 = G1Point.msm(points=points, scalars=scalars_msm).add(proof.shplonk_q)
    P_1 = -proof.kzg_quotient

    pairs = [G1G2Pair(P_0, G2_POINT_KZG_1), G1G2Pair(P_1, G2_POINT_KZG_2)]

    mpc_builder = MPCheckCalldataBuilder(
        curve_id=CurveID.BN254, pairs=pairs, n_fixed_g2=2, public_pair=None
    )
    cd = []
    cd.extend(proof.serialize_to_calldata())
    cd.extend(
        msm_builder.serialize_to_calldata(
            include_points_and_scalars=False,
            serialize_as_pure_felt252_array=True,
        )
    )
    cd.extend(mpc_builder.serialize_to_calldata())

    res = [len(cd)] + cd

    # print(f"HONK CALLDATA: {res}")
    # print(f"HONK CALLDATA LENGTH: {len(res)}")

    return res


def _honk_calldata_from_vk_and_proof_rust(
    vk: HonkVk,
    proof: Union[HonkProof, ZKHonkProof],
    system: ProofSystem = ProofSystem.UltraKeccakHonk,
) -> list[int]:
    match system:
        case ProofSystem.UltraKeccakHonk:
            flavor = 0
            zk = False
        case ProofSystem.UltraStarknetHonk:
            flavor = 1
            zk = False
        case ProofSystem.UltraKeccakZKHonk:
            flavor = 0
            zk = True
        case ProofSystem.UltraStarknetZKHonk:
            flavor = 1
            zk = True
        case _:
            raise ValueError(f"Proof system {system} not compatible")

    return garaga_rs.get_honk_calldata(
        proof.proof_bytes, proof.public_inputs_bytes, vk.vk_bytes, flavor, zk
    )



from garaga.algebra import PyFelt
from garaga.definitions import CURVES
from garaga.extension_field_modulo_circuit import ModuloCircuitElement
from garaga.modulo_circuit import ModuloCircuit, WriteOps
from garaga.precompiled_circuits.fp2 import Fp2Circuits


class MapToCurveG2(Fp2Circuits):
    def __init__(self, name: str, curve_id: int, compilation_mode: int = 0):
        super().__init__(
            name=name,
            curve_id=curve_id,
            compilation_mode=compilation_mode,
            generic_circuit=True,
        )
        self.curve = CURVES[curve_id]

    def set_consts(self):
        # Standard G2 curve parameters for BLS12-381
        # ToDo: Derive from definitions.py
        self.swu_a = [
            self.set_or_get_constant(0),  # real part
            self.set_or_get_constant(240),  # imaginary part
        ]

        self.swu_b = [
            self.set_or_get_constant(1012),  # real part
            self.set_or_get_constant(1012),  # imaginary part
        ]

        self.swu_z = [
            self.set_or_get_constant(-2),  # real part
            self.set_or_get_constant(-1),  # imaginary part
        ]

        self.one = [
            self.set_or_get_constant(1),  # real part
            self.set_or_get_constant(0),  # imaginary part
        ]

        self.zero = [
            self.set_or_get_constant(0),  # real part
            self.set_or_get_constant(0),  # imaginary part
        ]

    def map_to_curve_part_1(self, input_value: list[ModuloCircuitElement]):
        """
        Implements the first part of the Simplified SWU map-to-curve algorithm in G2.
        This maps a field element to a point on the curve E': y² = x³ + a*x + b.

        The algorithm follows these steps:
        1. Let u be the input field element
        2. Calculate intermediate values:
            - t = z²u⁴ + zu²  (where z is the SWU constant)
            - x₁ = (-b/a) * (1 + 1/(z²u⁴ + zu²))

        Args:
            input_value: Field element u in Fp2 to be mapped to the curve

        Returns:
            tuple: (g1x, div) where:
                - g1x is x³ + ax + b evaluated at the calculated x₁
                - div is an intermediate value needed for the full calculation
        """
        # Calculate u² and related terms
        u2 = self.fp2_square(input_value)  # u²
        zeta_u2 = self.fp2_mul(self.swu_z, u2)  # z·u²
        zeta_u2_square = self.fp2_square(zeta_u2)  # z²u⁴
        ta = self.fp2_add(zeta_u2_square, zeta_u2)  # t = z²u⁴ + zu²
        neg_ta = self.fp2_sub(self.zero, ta)  # -t
        num_x1 = self.fp2_mul(self.swu_b, self.fp2_add(ta, self.one))  # b(t + 1)

        # Handle special case when t = 0
        is_non_zero = self.fp2_is_non_zero(neg_ta)  # 1 if t ≠ 0, 0 if t = 0
        neg_ta_or_z = self.fp2_add(
            self.fp2_mul(self.fp2_sub(self.one, is_non_zero), self.swu_z),  # z if t = 0
            self.fp2_mul(is_non_zero, neg_ta),  # -t if t ≠ 0
        )

        # Calculate x₁ numerator and denominator
        div = self.fp2_mul(self.swu_a, neg_ta_or_z)  # a·(-t) or a·z
        num2_x1 = self.fp2_square(num_x1)  # (b(t + 1))²
        div2 = self.fp2_square(div)  # (a·(-t))²
        div3 = self.fp2_mul(div, div2)  # (a·(-t))³

        # Calculate g(x₁) = x₁³ + ax₁ + b
        num_gx1 = self.fp2_add(
            self.fp2_mul(
                self.fp2_add(num2_x1, self.fp2_mul(self.swu_a, div2)),  # x₁²  # ax₁
                num_x1,  # ·x₁ (completing the x₁³ term)
            ),
            self.fp2_mul(self.swu_b, div3),  # + b
        )
        g1x = self.fp2_mul(num_gx1, self.fp2_inv(div3))  # Final result normalized

        return [g1x, div, num_x1, zeta_u2]

    def compute_initial_coordinates_quadratic(
        self,
        field: list[ModuloCircuitElement],
        g1x: list[ModuloCircuitElement],
        div: list[ModuloCircuitElement],
        num_x1: list[ModuloCircuitElement],
    ):
        """
        Computes the initial coordinates for the map-to-curve operation when g1x is a
        quadratic residue. This function computes the initial y-coordinate before sign
        adjustment.

        IMPORTANT: This function should only be called when g1x is a quadratic residue,
        meaning there exists a y such that y² = g1x in the field Fp2.
        If g1x is not a quadratic residue, the fp2_sqrt operation will fail.

        The algorithm follows these steps:
        1. Compute y = √(g1x) where g1x = x³ + ax + b
           (requires g1x to be a quadratic residue)
        2. Compute x = num_x1/div to get the x-coordinate in affine form

        Args:
            field: The original input field element
            g1x: The value x³ + ax + b from the first part (must be a quadratic residue)
            div: The denominator from the first part
            num_x1: The numerator for x from the first part

        Returns:
            list: [x_affine, y_initial, field] where:
                - x_affine: The final x-coordinate
                - y_initial: The initial y-coordinate (needs sign adjustment)
                - field: The original field element (needed for sign adjustment)

        Note:
            The returned y_initial needs to be adjusted using adjust_y_sign() to ensure
            it has the correct sign relative to the input field element.
        """
        # Compute y-coordinate as the square root of g1x
        # This will only work if g1x is a quadratic residue
        y = self.fp2_sqrt(g1x)

        # Convert x to affine coordinates
        x_affine = self.fp2_div(num_x1, div)

        return [x_affine, y, field]

    def compute_initial_coordinates_non_quadratic(
        self,
        field: list[ModuloCircuitElement],
        g1x: list[ModuloCircuitElement],
        div: list[ModuloCircuitElement],
        num_x1: list[ModuloCircuitElement],
        zeta_u2: list[ModuloCircuitElement],
    ):
        """
        Computes the initial coordinates for the map-to-curve operation when g1x is NOT
        a quadratic residue. This function uses a clever mathematical property to compute
        a valid initial y-coordinate when direct square root is impossible.

        Key Mathematical Insight:
        When g1x is not a quadratic residue, we use the SWU constant z (which is also
        not a quadratic residue) to compute the y-coordinate. This works because:
        1. If g1x is not a quadratic residue, then z·g1x IS a quadratic residue
           (product of two non-quadratic residues is a quadratic residue)
        2. We can then compute y₁ = √(z·g1x)
        3. The initial y is computed as y = zu·field·y₁

        The algorithm:
        1. Compute y₁ = √(z·g1x)  [This is possible because z·g1x is a quadratic residue]
        2. Compute initial y = zu·field·y₁
        3. Compute x = (zu·num_x1)/div

        Args:
            field: The original input field element
            g1x: The value x³ + ax + b (known to be a non-quadratic residue)
            zeta_u2: The value z·u² from the first part
            num_x1: The x numerator from the first part
            div: The denominator from the first part

        Returns:
            list: [x_affine, y_initial, field] where:
                - x_affine: The final x-coordinate
                - y_initial: The initial y-coordinate (needs sign adjustment)
                - field: The original field element (needed for sign adjustment)

        Note:
            The returned y_initial needs to be adjusted using adjust_y_sign() to ensure
            it has the correct sign relative to the input field element.
        """
        # Since z·g1x is a quadratic residue (product of two non-quadratic residues),
        # this square root is guaranteed to exist
        y1 = self.fp2_sqrt(self.fp2_mul(self.swu_z, g1x))

        # Compute initial y-coordinate
        y = self.fp2_mul(zeta_u2, self.fp2_mul(field, y1))

        # Compute x-coordinate in affine form
        num_x = self.fp2_mul(zeta_u2, num_x1)
        x_affine = self.fp2_div(num_x, div)

        return [x_affine, y, field]

    def adjust_y_sign(
        self,
        y: list[ModuloCircuitElement],
        field: list[ModuloCircuitElement],
    ):
        # Handle sign adjustment as before
        qy, y_parity = self.fp2_parity(y)
        qfield, element_parity = self.fp2_parity(field)

        # Compute if parities are the same using XNOR (opposite of XOR)
        # XNOR(a,b) = 1 - (a + b - 2ab) = 2ab - a - b + 1
        same_parity = [
            self.add(
                self.sub(
                    self.mul(
                        self.set_or_get_constant(2),
                        self.mul(y_parity[0], element_parity[0]),
                    ),
                    self.add(y_parity[0], element_parity[0]),
                ),
                self.set_or_get_constant(1),
            ),
            self.zero[0],  # imaginary part is 0
        ]
        # Adjust y sign if parities don't match:
        # y_affine = same_parity ? y : -y
        y_affine = self.fp2_add(
            self.fp2_mul([same_parity[0], same_parity[1]], y),  # Keep y if same parity
            self.fp2_mul(
                [
                    self.sub(self.one[0], same_parity[0]),
                    self.zero[0],
                ],  # [1 - same_parity, 0]
                self.fp2_sub(self.zero, y),  # -y
            ),
        )

        return [y_affine, qfield, qy]

    # ATTENTION: this function is not sound by itself!
    # Two solutions are possible:
    # 1. (q, r)
    # 2. (q+ (p-1)/2, 1-r)
    # To constraint this properly, we still need assert: q <= (p-1)/2
    # This needs to happen in the function wrapping the garaga circuit
    def fp2_parity(
        self, element: list[ModuloCircuitElement]
    ) -> list[ModuloCircuitElement]:
        """
        Computes the parity of the first non-zero coefficient of the Fp2 element (element[0], element[1])
        Returns [parity, 0]

        For an Fp2 element a + bi:
        1. If a ≠ 0, returns parity of a
        2. If a = 0 and b ≠ 0, returns parity of b
        3. If both are 0, returns [0, 0] (even)

        Implements sgn0_m_eq_2 from RFC9380 using witness variables for validation.
        """
        assert len(element) == 2 and all(
            isinstance(x, ModuloCircuitElement) for x in element
        )

        two = self.set_or_get_constant(2)
        one = self.set_or_get_constant(1)
        zero = self.set_or_get_constant(0)

        # For element[0] (real part)
        # Witnesses: q0 (quotient), r0 (remainder)
        q0 = self.write_element(
            PyFelt(element[0].value // 2, element[0].p), WriteOps.WITNESS
        )  # Witness for q0
        r0 = self.write_element(
            PyFelt(element[0].value % 2, element[0].p), WriteOps.WITNESS
        )  # Witness for r0 (parity of x0)

        # Enforce that r0 ∈ {0, 1}
        r0_sub_1 = self.sub(r0, one)
        r0_times_r0_sub_1 = self.mul(r0, r0_sub_1)
        self.sub_and_assert(r0_times_r0_sub_1, zero, zero, comment="Ensure r0 ∈ {0,1}")

        # Enforce x0 = 2 * q0 + r0
        two_q0 = self.mul(q0, two)
        self.add_and_assert(two_q0, r0, element[0], comment="Validate x0 decomposition")

        # Similarly for element[1] (imaginary part)
        q1 = self.write_element(
            PyFelt(element[1].value // 2, element[1].p), WriteOps.WITNESS
        )  # Witness for q1
        r1 = self.write_element(
            PyFelt((element[1].value % 2), element[1].p), WriteOps.WITNESS
        )  # Witness for r1 (parity of x1)

        # Enforce that r1 ∈ {0, 1}
        r1_sub_1 = self.sub(r1, one)
        r1_times_r1_sub_1 = self.mul(r1, r1_sub_1)
        self.sub_and_assert(r1_times_r1_sub_1, zero, zero, comment="Ensure r1 ∈ {0,1}")

        # Enforce x1 = 2 * q1 + r1
        two_q1 = self.mul(q1, two)
        self.add_and_assert(two_q1, r1, element[1], comment="Validate x1 decomposition")

        # Compute zero_0 = 1 - fp_is_non_zero(x0)
        real_is_non_zero = self.fp_is_non_zero(
            element[0]
        )  # Returns 1 if x0 ≠ 0, else 0
        zero_0 = self.sub(one, real_is_non_zero)

        # Compute s = r0 OR (zero_0 AND r1)
        # Implementing logical operations using arithmetic operations
        # zero_0 AND r1
        zero_0_and_r1 = self.mul(zero_0, r1)

        # r0 OR (zero_0 AND r1)
        # OR(a, b) = a + b - a * b
        or_input = self.sub(self.add(r0, zero_0_and_r1), self.mul(r0, zero_0_and_r1))
        s = or_input  # Since values are in {0,1}, this computes the logical OR

        # select q based on s
        q = self.add(
            self.mul(q0, self.sub(one, zero_0_and_r1)),  # q0 when not using r1
            self.mul(q1, zero_0_and_r1),  # q1 when using r1
        )

        # Return parity as [s, 0]
        return q, [s, zero]


class MapToCurveG1(ModuloCircuit):
    def __init__(self, name: str, curve_id: int, compilation_mode: int = 0):
        super().__init__(
            name=name,
            curve_id=curve_id,
            compilation_mode=compilation_mode,
            generic_circuit=True,
        )
        self.curve = CURVES[curve_id]

    def set_consts(self):
        params = self.curve.swu_params
        self.swu_a = self.set_or_get_constant(params.A)
        self.swu_b = self.set_or_get_constant(params.B)
        self.swu_z = self.set_or_get_constant(params.Z)
        self.zero = self.set_or_get_constant(0)
        self.one = self.set_or_get_constant(1)

    def map_to_curve_part_1(self, input_value: ModuloCircuitElement):
        """
        Implements the first part of the Simplified SWU map-to-curve algorithm for G1.
        This maps a field element to a point on the curve E': y² = x³ + ax + b.

        The algorithm follows these steps:
        1. Let u be the input field element
        2. Calculate t = z²u⁴ + zu² (where z is the SWU constant)
        3. Calculate x₁ = (-b/a) * (1 + 1/t) if t ≠ 0, or x₁ = (-b/(az)) if t = 0
        4. Calculate g(x₁) = x₁³ + ax₁ + b

        Args:
            input_value (ModuloCircuitElement): Field element u to be mapped to the curve

        Returns:
            list: [g1x, div, num_x1, zeta_u2] where:
                - g1x: The value g(x₁) = x₁³ + ax₁ + b
                - div: The denominator used in x₁ calculation (needed for finalization)
                - num_x1: The numerator of x₁ (needed for finalization)
                - zeta_u2: The value z·u² (needed for non-quadratic case)

        Note:
            This is part 1 of the map-to-curve operation. The result needs to be finalized
            using either finalize_map_to_curve_quadratic or finalize_map_to_curve_non_quadratic,
            depending on whether g1x is a quadratic residue.
        """
        # Calculate u² and related terms
        u2 = self.square(input_value)  # u²
        zeta_u2 = self.mul(self.swu_z, u2)  # z·u²
        zeta_u2_square = self.square(zeta_u2)  # z²u⁴
        ta = self.add(zeta_u2_square, zeta_u2)  # t = z²u⁴ + zu²
        neg_ta = self.sub(self.zero, ta)  # -t
        num_x1 = self.mul(self.swu_b, self.add(ta, self.one))  # b(t + 1)

        # Handle special case when t = 0
        is_non_zero = self.fp_is_non_zero(neg_ta)  # 1 if t ≠ 0, 0 if t = 0
        neg_ta_or_z = self.add(
            self.mul(self.sub(self.one, is_non_zero), self.swu_z),  # z if t = 0
            self.mul(is_non_zero, neg_ta),  # -t if t ≠ 0
        )

        # Calculate x₁ numerator and denominator
        div = self.mul(self.swu_a, neg_ta_or_z)  # a·(-t) or a·z
        num2_x1 = self.square(num_x1)  # (b(t + 1))²
        div2 = self.square(div)  # (a·(-t))²
        div3 = self.mul(div, div2)  # (a·(-t))³

        # Calculate g(x₁) = x₁³ + ax₁ + b
        num_gx1 = self.add(
            self.mul(
                self.add(num2_x1, self.mul(self.swu_a, div2)),  # x₁²  # ax₁
                num_x1,  # ·x₁ (completing the x₁³ term)
            ),
            self.mul(self.swu_b, div3),  # + b
        )
        g1x = self.mul(num_gx1, self.inv(div3))  # Final result normalized

        return [g1x, div, num_x1, zeta_u2]

    def compute_initial_coordinates_quadratic(
        self,
        field: ModuloCircuitElement,
        g1x: ModuloCircuitElement,
        div: ModuloCircuitElement,
        num_x1: ModuloCircuitElement,
    ):
        """
        Computes the x and y coordinates for the map-to-curve operation when g1x is a
        quadratic residue. This function computes the y-coordinate and ensures the point
        has the correct sign.

        IMPORTANT: This function should only be called when g1x is a quadratic residue,
        meaning there exists a y such that y² = g1x in the field.

        The algorithm follows these steps:
        1. Compute y = √(g1x) where g1x = x³ + ax + b
           (requires g1x to be a quadratic residue)
        2. Compute x = num_x1/div to get the x-coordinate in affine form
        3. Adjust the sign of y to match the parity of the input field element

        The sign adjustment uses:
            - If sign(y) ≠ sign(field_element): y = -y
            - For field elements, parity is determined by the real part

        Args:
            field: The original input field element
            g1x: The value x³ + ax + b from the first part (must be a quadratic residue)
            div: The denominator from the first part
            num_x1: The numerator for x from the first part

        Returns:
            list: [x_affine, y_initial, field] where:
                - x_affine: The final x-coordinate
                - y_initial: The initial y-coordinate (needs sign adjustment)
                - field: The original field element (needed for sign adjustment)

        Note:
            When g1x is not a quadratic residue, the alternative function
            compute_initial_coordinates_non_quadratic should be used instead.
        """
        # Compute y-coordinate as the square root of g1x
        y_initial = self.fp_sqrt(g1x)

        # Convert x to affine coordinates
        x_affine = self.div(num_x1, div)

        return [x_affine, y_initial, field]

    def compute_initial_coordinates_non_quadratic(
        self,
        field: ModuloCircuitElement,
        g1x: ModuloCircuitElement,
        div: ModuloCircuitElement,
        num_x1: ModuloCircuitElement,
        zeta_u2: ModuloCircuitElement,
    ):
        """
        Computes the initial x and y coordinates for the map-to-curve operation when g1x
        is NOT a quadratic residue. The y-coordinate will need further sign adjustment
        using adjust_y_sign().

        Key Mathematical Insight:
        When g1x is not a quadratic residue, we use the SWU constant z (which is also
        not a quadratic residue) to compute the y-coordinate. This works because:
        1. If g1x is not a quadratic residue, then z·g1x IS a quadratic residue
           (product of two non-quadratic residues is a quadratic residue)
        2. We can then compute y₁ = √(z·g1x)
        3. The initial y is computed as y = zu·field·y₁

        The algorithm:
        1. Compute y₁ = √(z·g1x)  [This is possible because z·g1x is a quadratic residue]
        2. Compute initial y = zu·field·y₁
        3. Compute x = (zu·num_x1)/div

        Args:
            field: The original input field element
            g1x: The value x³ + ax + b (known to be a non-quadratic residue)
            zeta_u2: The value z·u² from the first part
            num_x1: The x numerator from the first part
            div: The denominator from the first part

        Returns:
            list: [x_affine, y_initial, field] where:
                - x_affine: The final x-coordinate
                - y_initial: The initial y-coordinate (needs sign adjustment)
                - field: The original field element (needed for sign adjustment)

        Note:
            The returned y_initial needs to be adjusted using adjust_y_sign() to ensure
            it has the correct sign relative to the input field element.
        """
        # Since z·g1x is a quadratic residue (product of two non-quadratic residues),
        # this square root is guaranteed to exist
        y1 = self.fp_sqrt(self.mul(self.swu_z, g1x))

        # Compute initial y-coordinate
        y_initial = self.mul(zeta_u2, self.mul(field, y1))

        # Compute x-coordinate in affine form
        num_x = self.mul(zeta_u2, num_x1)
        x_affine = self.div(num_x, div)

        return [x_affine, y_initial, field]

    def adjust_y_sign(
        self,
        y: ModuloCircuitElement,
        field: ModuloCircuitElement,
    ):
        # Handle sign adjustment as before
        qy, y_parity = self.fp_parity(y)
        qfield, element_parity = self.fp_parity(field)

        # Compute if parities are the same using XNOR (opposite of XOR)
        # XNOR(a,b) = 1 - (a + b - 2ab) = 2ab - a - b + 1
        same_parity = self.add(
            self.sub(
                self.mul(
                    self.set_or_get_constant(2),
                    self.mul(y_parity, element_parity),
                ),
                self.add(y_parity, element_parity),
            ),
            self.set_or_get_constant(1),
        )

        # Adjust y sign if parities don't match:
        # y_affine = same_parity ? y : -y
        y_affine = self.add(
            self.mul(same_parity, y),  # Keep y if same parity
            self.mul(
                self.sub(self.one, same_parity),
                self.sub(self.zero, y),  # -y
            ),
        )

        return [y_affine, qy, qfield]

    # ATTENTION: this function is not sound by itself!
    # Two solutions are possible:
    # 1. (q, r)
    # 2. (q+ (p-1)/2, 1-r)
    # To constraint this properly, we still need assert: q <= (p-1)/2
    # This needs to happen in the function wrapping the garaga circuit
    def fp_parity(self, element: ModuloCircuitElement) -> ModuloCircuitElement:
        """
        Computes the parity of a field element.
        Returns 0 if element is even, 1 if odd.

        Implements sgn0_m_eq_1 from RFC9380 using witness variables for validation.
        """
        assert isinstance(element, ModuloCircuitElement)

        two = self.set_or_get_constant(2)
        one = self.set_or_get_constant(1)
        zero = self.set_or_get_constant(0)

        # Witnesses: q (quotient), r (remainder)
        q = self.write_element(
            PyFelt(element.value // 2, element.p), WriteOps.WITNESS
        )  # Witness for quotient
        r = self.write_element(
            PyFelt(element.value % 2, element.p), WriteOps.WITNESS
        )  # Witness for remainder (parity)

        # Enforce that r ∈ {0, 1}
        r_sub_1 = self.sub(r, one)
        r_times_r_sub_1 = self.mul(r, r_sub_1)
        self.sub_and_assert(r_times_r_sub_1, zero, zero, comment="Ensure r ∈ {0,1}")

        # Enforce element = 2 * q + r
        two_q = self.mul(q, two)
        self.add_and_assert(two_q, r, element, comment="Validate element decomposition")

        return q, r


if __name__ == "__main__":
    circuit = MapToCurveG1("map_to_curve", 1)  # BLS12-381
    circuit.set_consts()

    field = circuit.write_element(
        circuit.field(
            2231413721970278425038638834062370180699174210864795385441649994565282274875534254514118105433862522641883533654145
        )
    )

    g1x, div, num_x1, zeta_u2 = circuit.map_to_curve_part_1(field)

    print("g1x", g1x)
    print("div", div)
    print("num_x1", num_x1)
    print("zeta_u2", zeta_u2)

    if g1x.felt.is_quad_residue():
        print("Quadratic residue")
        (x_affine, y_initial, field) = circuit.compute_initial_coordinates_quadratic(
            field, g1x, div, num_x1
        )
    else:
        print("Non quadratic residue")
        (x_affine, y_initial, field) = (
            circuit.compute_initial_coordinates_non_quadratic(
                field, g1x, div, num_x1, zeta_u2
            )
        )

    # Adjust y sign
    (y_affine, qy, qfield) = circuit.adjust_y_sign(field, y_initial)

    print("x_affine", x_affine)
    print("y_affine", y_affine)



from abc import ABC, abstractmethod
from typing import Dict, List, Optional, Tuple, Type, Union

import garaga.modulo_circuit_structs as structs
from garaga.definitions import BLS12_381_ID, BN254_ID, get_irreducible_poly
from garaga.extension_field_modulo_circuit import ExtensionFieldModuloCircuit, PyFelt
from garaga.modulo_circuit import ModuloCircuit
from garaga.modulo_circuit_structs import (
    E12D,
    BLSProcessedPair,
    BNProcessedPair,
    Cairo1SerializableStruct,
    G1PointCircuit,
    G2Line,
    G2PointCircuit,
    MillerLoopResultScalingFactor,
    u384,
    u384Array,
)
from garaga.precompiled_circuits import multi_pairing_check
from garaga.precompiled_circuits.compilable_circuits.base import BaseEXTFCircuit


def split_4_sized_object_into_tuple_of_2_size(
    input: Union[List[PyFelt], Tuple[PyFelt, PyFelt, PyFelt, PyFelt]]
) -> Optional[Tuple[List[PyFelt], List[PyFelt]]]:
    if input is None:
        return None
    assert len(input) == 4, f"Expected input of length 4, got {len(input)}"
    return (list(input[0:2]), list(input[2:4]))


def parse_precomputed_g1_consts_and_g2_points(
    circuit: multi_pairing_check.MultiPairingCheckCircuit,
    vars: dict,
    n_pairs: int,
    bit_1: bool = False,
) -> list[tuple[list[PyFelt], list[PyFelt]]]:
    current_points = []
    q_or_q_neg_points = [] if bit_1 else None
    for i in range(n_pairs):
        circuit.yInv.append(vars[f"yInv_{i}"])
        circuit.xNegOverY.append(vars[f"xNegOverY_{i}"])
        current_points.append(
            split_4_sized_object_into_tuple_of_2_size(vars.get(f"Q_{i}", None))
        )  # Return empty list if not present
        if bit_1:
            q_or_q_neg_points.append(
                split_4_sized_object_into_tuple_of_2_size(
                    vars.get(f"Q_or_Q_neg_{i}", None)
                )  # Return empty list if not present
            )

    return current_points, q_or_q_neg_points


class BaseFixedG2PointsMPCheck(BaseEXTFCircuit, ABC):
    def __init__(
        self,
        name,
        curve_id,
        n_pairs: int,
        n_fixed_g2: int,
        auto_run=True,
        compilation_mode=1,
    ):
        assert compilation_mode == 1, "Compilation mode 1 is required for this circuit"
        if n_pairs and "mp_check_prepare_pairs" not in name:
            assert (
                2 <= n_pairs
            ), f"Multi-pairing check requires at least 2 pairs, got {n_pairs}"

        if n_fixed_g2:
            assert (
                0 <= n_fixed_g2 <= n_pairs
            ), f"Number of fixed G2 points must be between 0 and {n_pairs}, got {n_fixed_g2}"

        self.n_pairs = n_pairs
        self.n_fixed_g2 = n_fixed_g2
        super().__init__(
            name=name,
            curve_id=curve_id,
            auto_run=auto_run,
            compilation_mode=compilation_mode,
        )
        self.generic_over_curve = True

    def _initialize_circuit(self):
        return multi_pairing_check.MultiPairingCheckCircuit(
            self.name,
            self.curve_id,
            n_pairs=self.n_pairs,
            hash_input=False,
            compilation_mode=self.compilation_mode,
            precompute_lines=True,
            n_points_precomputed_lines=self.n_fixed_g2,
        )

    @property
    @abstractmethod
    def input_map(
        self,
    ) -> Dict[
        str,
        Union[
            Type[Cairo1SerializableStruct], Tuple[Type[Cairo1SerializableStruct], int]
        ],
    ]:
        """
        Define the input map for the circuit in a dict.
        The key will be the name of the input variable, also used in the signature of the compiled Cairo code.
        The value will be either a Cairo1SerializableStruct type (which defines the struct in the Cairo code),
            or a tuple of the type and its size (for Array-like types).
        The reason behind this is that each Cairo1SerializableStruct defines the __len__ method, but for the
        array-like structs we need to specify the size in advance.
        """

    def _base_input_map(self, bit_type: str) -> dict:
        """
        Base input map for the bit 0, 1, 00, 01, and 10 cases.
        """
        input_map = {}

        # Add pair inputs
        for k in range(self.n_fixed_g2):
            input_map[f"yInv_{k}"] = u384
            input_map[f"xNegOverY_{k}"] = u384
            input_map[f"G2_line_dbl_{k}"] = G2Line
            if bit_type in ("1"):
                input_map[f"G2_line_add_{k}"] = G2Line
            if bit_type == "10":
                input_map[f"G2_line_add_1_{k}"] = G2Line
                input_map[f"G2_line_dbl_0_{k}"] = G2Line
            if bit_type == "01":
                input_map[f"G2_line_dbl_1{k}"] = G2Line
                input_map[f"G2_line_add_1{k}"] = G2Line
            if bit_type == "00":
                input_map[f"G2_line_2nd_0_{k}"] = G2Line

        for k in range(self.n_fixed_g2, self.n_pairs):
            input_map[f"yInv_{k}"] = u384
            input_map[f"xNegOverY_{k}"] = u384
            input_map[f"Q_{k}"] = G2PointCircuit
            if bit_type in ("1", "01", "10"):
                input_map[f"Q_or_Q_neg_{k}"] = G2PointCircuit

        # Add common inputs
        input_map["lhs_i"] = u384
        input_map["f_i_of_z"] = u384
        input_map["f_i_plus_one_of_z"] = u384

        # Add bit-specific inputs
        if bit_type in ("1", "01", "10"):
            input_map["c_or_cinv_of_z"] = u384

        input_map["z"] = u384
        input_map["ci"] = u384

        return input_map

    def _process_input(
        self, circuit: multi_pairing_check.MultiPairingCheckCircuit, input: list[PyFelt]
    ) -> dict:
        """
        Method responsible for deserializing the input list of elements into the variables in the input map,
        and writing them to the circuit.
        The input list is expected to be in the same order as the input map.
        Since we use Python 3.10, the input map dict is ordered.
        Returns a vars dict with the same keys as the input map, but with the values being the instances of the structs,
        each struct holding ModuloCircuitElement(s).
        """
        vars = {}
        for name, struct_info in self.input_map.items():
            if isinstance(struct_info, tuple) and struct_info[0] == u384Array:
                # Handle u384Array with specified size
                struct_type, size = struct_info
                vars[name] = circuit.write_struct(
                    struct_type(name, elmts=[input.pop(0) for _ in range(size)])
                )
            elif struct_info == G2Line:
                # Internally appended to the precomputed_lines flattened list
                circuit.precomputed_lines.extend(
                    circuit.write_struct(
                        G2Line(name, elmts=[input.pop(0) for _ in range(4)])
                    )
                )
            else:
                struct_type = struct_info
                # For other types, create a temporary instance to get its length
                temp_instance = struct_type(
                    name, None
                )  # Temporary instance with a dummy element
                size = len(temp_instance)
                if size == 1:
                    vars[name] = circuit.write_struct(
                        struct_type(name, elmts=[input.pop(0)])
                    )
                else:
                    # Convert to tuple so that it is Hashable
                    vars[name] = tuple(
                        circuit.write_struct(
                            struct_type(name, elmts=[input.pop(0) for _ in range(size)])
                        )
                    )
        assert len(input) == 0, f"Expected input of length 0, got {len(input)}"
        # Create the precomputed lines generator
        circuit._precomputed_lines_generator = (
            circuit._create_precomputed_lines_generator()
        )
        return vars

    def build_input(self) -> list[PyFelt]:
        """
        Extends the base method of BaseModuloCircuit, by reading the input map and returning a list of random elements of the total expected size.
        """
        total_elements = 0
        for name, struct_info in self.input_map.items():
            if isinstance(struct_info, tuple):
                # Array-like case
                _, size = struct_info
                total_elements += size
            else:
                # For other structs, create a temporary instance to get its length
                temp_instance = struct_info(name, None)
                total_elements += len(temp_instance)

        return [self.field.random() for _ in range(total_elements)]

    def _run_circuit_inner(self, input: list[PyFelt]) -> ModuloCircuit:
        circuit = self._initialize_circuit()

        vars = self._process_input(circuit, input)
        return self._execute_circuit_logic(circuit, vars)

    @abstractmethod
    def _execute_circuit_logic(self, circuit, vars) -> ModuloCircuit:
        """
        Implement the circuit logic using the processed input variables.
        """

    def _execute_circuit_bit_logic_base(self, circuit: ModuloCircuit, vars, bit_type):
        n_pairs = self.n_pairs
        assert n_pairs >= 2, f"n_pairs must be >= 2, got {n_pairs}"

        current_points, q_or_q_neg_points = parse_precomputed_g1_consts_and_g2_points(
            circuit,
            vars,
            n_pairs,
            bit_1=(bit_type == "1" or bit_type == "01" or bit_type == "10"),
        )

        circuit.create_lines_z_powers(vars["z"])
        ci_plus_one = circuit.mul(vars["ci"], vars["ci"], "Compute c_i = (c_(i-1))^2")

        sum_i_prod_k_P = circuit.mul(
            vars["f_i_of_z"],
            vars["f_i_of_z"],
            "Square f evaluation in Z, the result of previous bit.",
        )

        new_points, sum_i_prod_k_P = self._process_points(
            circuit, current_points, q_or_q_neg_points, sum_i_prod_k_P, bit_type
        )

        if bit_type in ("1", "01"):
            sum_i_prod_k_P = circuit.mul(sum_i_prod_k_P, vars["c_or_cinv_of_z"])
        elif bit_type == "10":
            sum_i_prod_k_P = circuit.mul(
                sum_i_prod_k_P, circuit.square(vars["c_or_cinv_of_z"])
            )

        f_i_plus_one_of_z = vars["f_i_plus_one_of_z"]
        new_lhs = circuit.mul(
            ci_plus_one,
            circuit.sub(sum_i_prod_k_P, f_i_plus_one_of_z, "(Π(i,k) (Pk(z))) - Ri(z)"),
            "ci * ((Π(i,k) (Pk(z)) - Ri(z))",
        )
        lhs_i_plus_one = circuit.add(
            vars["lhs_i"], new_lhs, "LHS = LHS + ci * ((Π(i,k) (Pk(z)) - Ri(z))"
        )

        self._extend_output(circuit, new_points, lhs_i_plus_one, ci_plus_one)

        return circuit

    def _process_points(
        self,
        circuit: multi_pairing_check.MultiPairingCheckCircuit,
        current_points,
        q_or_q_neg_points,
        sum_i_prod_k_P,
        bit_type,
    ):
        new_points = []
        if bit_type == "00":
            for k in range(self.n_pairs):
                T1, l1 = circuit.double_step(current_points[k], k)
                sum_i_prod_k_P = self._multiply_line_evaluations(
                    circuit, sum_i_prod_k_P, [l1], k
                )
                new_points.append(T1)

            sum_i_prod_k_P = circuit.mul(
                sum_i_prod_k_P,
                sum_i_prod_k_P,
                "Compute (f^2 * Π(i,k) (line_i,k(z))) ^ 2 = f^4 * (Π(i,k) (line_i,k(z)))^2",
            )

            new_new_points = []
            for k in range(self.n_pairs):
                T, l2 = circuit.double_step(new_points[k], k)
                sum_i_prod_k_P = self._multiply_line_evaluations(
                    circuit, sum_i_prod_k_P, [l2], k
                )
                new_new_points.append(T)
            return new_new_points, sum_i_prod_k_P
        elif bit_type == "01":
            for k in range(self.n_pairs):
                T, l1 = circuit.double_step(current_points[k], k)
                sum_i_prod_k_P = self._multiply_line_evaluations(
                    circuit, sum_i_prod_k_P, [l1], k
                )
                new_points.append(T)

            sum_i_prod_k_P = circuit.mul(
                sum_i_prod_k_P,
                sum_i_prod_k_P,
                "Compute (f^2 * Π(i,k) (line_i,k(z))) ^ 2 = f^4 * (Π(i,k) (line_i,k(z)))^2",
            )

            new_new_points = []
            for k in range(self.n_pairs):
                T, l1, l2 = circuit.double_and_add_step(
                    new_points[k], q_or_q_neg_points[k], k
                )
                sum_i_prod_k_P = self._multiply_line_evaluations(
                    circuit, sum_i_prod_k_P, [l1, l2], k
                )
                new_new_points.append(T)

            return new_new_points, sum_i_prod_k_P

        elif bit_type == "10":
            for k in range(self.n_pairs):
                T, l1, l2 = circuit.double_and_add_step(
                    current_points[k], q_or_q_neg_points[k], k
                )
                sum_i_prod_k_P = self._multiply_line_evaluations(
                    circuit, sum_i_prod_k_P, [l1, l2], k
                )
                new_points.append(T)

            sum_i_prod_k_P = circuit.mul(
                sum_i_prod_k_P,
                sum_i_prod_k_P,
                "Compute (f^2 * Π(i,k) (line_i,k(z))) ^ 2 = f^4 * (Π(i,k) (line_i,k(z)))^2",
            )
            new_new_points = []
            for k in range(self.n_pairs):
                T, l1 = circuit.double_step(new_points[k], k)
                sum_i_prod_k_P = self._multiply_line_evaluations(
                    circuit, sum_i_prod_k_P, [l1], k
                )
                new_new_points.append(T)
            return new_new_points, sum_i_prod_k_P
        elif bit_type == "0":
            for k in range(self.n_pairs):
                T, l1 = circuit.double_step(current_points[k], k)
                sum_i_prod_k_P = self._multiply_line_evaluations(
                    circuit, sum_i_prod_k_P, [l1], k
                )
                new_points.append(T)
        elif bit_type == "1":
            for k in range(self.n_pairs):
                T, l1, l2 = circuit.double_and_add_step(
                    current_points[k], q_or_q_neg_points[k], k
                )
                sum_i_prod_k_P = self._multiply_line_evaluations(
                    circuit, sum_i_prod_k_P, [l1, l2], k
                )
                new_points.append(T)

        else:
            raise ValueError(f"Invalid bit type: {bit_type}")
        return new_points, sum_i_prod_k_P

    def _multiply_line_evaluations(
        self,
        circuit: multi_pairing_check.MultiPairingCheckCircuit,
        sum_i_prod_k_P,
        lines,
        k,
    ):
        for i, l in enumerate(lines):
            sum_i_prod_k_P = circuit.mul(
                sum_i_prod_k_P,
                circuit.eval_poly_in_precomputed_Z(
                    l, circuit.line_sparsity, f"line_{k}p_{i+1}"
                ),
                f"Mul (f(z)^2 * Π_0_k-1(line_k(z))) * line_i_{k}(z)",
            )
        return sum_i_prod_k_P

    def _extend_output(self, circuit, new_points, lhs_i_plus_one, ci_plus_one):

        last_points = (
            new_points[-(self.n_pairs - self.n_fixed_g2) :]
            if self.n_pairs != self.n_fixed_g2
            else []
        )
        for i, point in enumerate(last_points):
            circuit.extend_struct_output(
                G2PointCircuit(
                    name=f"Q{i}",
                    elmts=[
                        point[0][0],
                        point[0][1],
                        point[1][0],
                        point[1][1],
                    ],
                )
            )
        circuit.extend_struct_output(
            u384(name="lhs_i_plus_one", elmts=[lhs_i_plus_one])
        )
        circuit.extend_struct_output(u384(name="ci_plus_one", elmts=[ci_plus_one]))


class FixedG2MPCheckBitBase(BaseFixedG2PointsMPCheck):
    """Base class for bit checking circuits with default parameters."""

    BIT_TYPE = None  # Override in subclasses
    DEFAULT_PAIRS = 3
    DEFAULT_FIXED_G2 = 2

    def __init__(
        self,
        curve_id: int,
        n_pairs: int = None,
        n_fixed_g2: int = None,
        auto_run: bool = True,
        compilation_mode: int = 1,
    ):
        assert compilation_mode == 1, "Compilation mode 1 is required for this circuit"
        n_pairs = n_pairs if n_pairs is not None else self.DEFAULT_PAIRS
        n_fixed_g2 = n_fixed_g2 if n_fixed_g2 is not None else self.DEFAULT_FIXED_G2

        super().__init__(
            name=f"mp_check_bit{self.BIT_TYPE}_{n_pairs}P_{n_fixed_g2}F",
            curve_id=curve_id,
            n_pairs=n_pairs,
            n_fixed_g2=n_fixed_g2,
            auto_run=auto_run,
            compilation_mode=compilation_mode,
        )

    @property
    def input_map(self):
        return self._base_input_map(self.BIT_TYPE)

    def _execute_circuit_logic(self, circuit, vars) -> ModuloCircuit:
        return self._execute_circuit_bit_logic_base(circuit, vars, self.BIT_TYPE)


class FixedG2MPCheckBit0(FixedG2MPCheckBitBase):
    BIT_TYPE = "0"


class FixedG2MPCheckBit00(FixedG2MPCheckBitBase):
    BIT_TYPE = "00"


class FixedG2MPCheckBit1(FixedG2MPCheckBitBase):
    BIT_TYPE = "1"


class FixedG2MPCheckBit01(FixedG2MPCheckBitBase):
    BIT_TYPE = "01"


class FixedG2MPCheckBit10(FixedG2MPCheckBitBase):
    BIT_TYPE = "10"


class FixedG2MPCheckInitBit(BaseFixedG2PointsMPCheck):
    def __init__(
        self,
        curve_id: int,
        auto_run: bool = True,
        compilation_mode: int = 1,
        n_pairs: int = 3,
        n_fixed_g2: int = 2,
    ):
        super().__init__(
            name=f"mp_check_init_bit_{n_pairs}P_{n_fixed_g2}F",
            curve_id=curve_id,
            n_pairs=n_pairs,
            n_fixed_g2=n_fixed_g2,
            auto_run=auto_run,
            compilation_mode=compilation_mode,
        )

    @property
    def input_map(self):
        input_map = {}
        for i in range(self.n_fixed_g2):
            input_map[f"yInv_{i}"] = u384
            input_map[f"xNegOverY_{i}"] = u384
            input_map[f"G2_line_{i}"] = G2Line
            if self.curve_id == BLS12_381_ID:
                input_map[f"G2_line_{i}_2"] = G2Line

        for i in range(self.n_fixed_g2, self.n_pairs):
            input_map[f"yInv_{i}"] = u384
            input_map[f"xNegOverY_{i}"] = u384
            input_map[f"Q_{i}"] = G2PointCircuit

        input_map.update(
            {
                "R_i_of_z": u384,
                "c0": u384,
                "z": u384,
                "c_inv_of_z": u384,
            }
        )
        if self.curve_id == BN254_ID:
            input_map["previous_lhs"] = u384
        return input_map

    def _execute_circuit_logic(self, circuit, vars) -> ModuloCircuit:
        n_pairs = self.n_pairs
        current_points, _ = parse_precomputed_g1_consts_and_g2_points(
            circuit, vars, n_pairs
        )

        c0 = vars["c0"]
        z = vars["z"]
        c_inv_of_z = vars["c_inv_of_z"]
        circuit.create_lines_z_powers(z)

        f_i_plus_one_of_z = vars["R_i_of_z"]

        sum_i_prod_k_P_of_z = circuit.mul(
            c_inv_of_z, c_inv_of_z
        )  # At initialisation, f=1/c so f^2 = 1/c^2
        new_points = []
        if self.curve_id == BN254_ID:
            c_i = circuit.mul(
                c0, c0
            )  # Second relation for BN at init bit, need to update c_i.
            for k in range(n_pairs):
                T, l1 = circuit.double_step(current_points[k], k)
                sum_i_prod_k_P_of_z = circuit.mul(
                    sum_i_prod_k_P_of_z,
                    circuit.eval_poly_in_precomputed_Z(
                        l1, circuit.line_sparsity, f"line_{k}p_1"
                    ),
                )
                new_points.append(T)
        elif self.curve_id == BLS12_381_ID:
            c_i = c0  # first relation for BLS at init bit, no need to update c_i.
            # bit +1:  multiply f^2 by 1/c
            sum_i_prod_k_P_of_z = circuit.mul(c_inv_of_z, sum_i_prod_k_P_of_z)
            for k in range(n_pairs):
                T, l1, l2 = circuit.triple_step(current_points[k], k)
                sum_i_prod_k_P_of_z = circuit.mul(
                    sum_i_prod_k_P_of_z,
                    circuit.eval_poly_in_precomputed_Z(
                        l1, circuit.line_sparsity, f"line_{k}p_1"
                    ),
                )
                sum_i_prod_k_P_of_z = circuit.mul(
                    sum_i_prod_k_P_of_z,
                    circuit.eval_poly_in_precomputed_Z(
                        l2, circuit.line_sparsity, f"line_{k}p_2"
                    ),
                )
                new_points.append(T)

        new_lhs = circuit.mul(
            c_i,
            circuit.sub(sum_i_prod_k_P_of_z, f_i_plus_one_of_z),
            comment="ci * ((Π(i,k) (Pk(z)) - Ri(z))",
        )

        if self.curve_id == BLS12_381_ID:
            new_lhs = new_lhs
        elif self.curve_id == BN254_ID:
            previous_lhs = vars["previous_lhs"]
            new_lhs = circuit.add(new_lhs, previous_lhs)

        # OUTPUT
        last_points = (
            new_points[-(self.n_pairs - self.n_fixed_g2) :]
            if self.n_fixed_g2 != self.n_pairs
            else []
        )
        for i, point in enumerate(last_points):
            circuit.extend_struct_output(
                G2PointCircuit(
                    name=f"Q{i}",
                    elmts=[
                        point[0][0],
                        point[0][1],
                        point[1][0],
                        point[1][1],
                    ],
                )
            )
        circuit.extend_struct_output(u384("new_lhs", elmts=[new_lhs]))
        if self.curve_id == BN254_ID:
            circuit.extend_struct_output(u384("c_i", elmts=[c_i]))

        return circuit


class FixedG2MPCheckFinalizeBN(BaseFixedG2PointsMPCheck):
    def __init__(
        self,
        curve_id: int,
        auto_run: bool = True,
        n_pairs: int = 3,
        n_fixed_g2: int = 2,
        compilation_mode: int = 1,
    ):
        self.max_q_degree = multi_pairing_check.get_max_Q_degree(curve_id, n_pairs)

        super().__init__(
            name=f"mp_check_finalize_bn_{n_pairs}P_{n_fixed_g2}F",
            curve_id=curve_id,
            n_pairs=n_pairs,
            n_fixed_g2=n_fixed_g2,
            auto_run=auto_run,
            compilation_mode=compilation_mode,
        )

    @property
    def input_map(self):
        if self.curve_id == BLS12_381_ID:
            return {}
        input_map = {}
        for i in range(self.n_fixed_g2):
            input_map[f"yInv_{i}"] = u384
            input_map[f"xNegOverY_{i}"] = u384
            input_map[f"line_1_{i}"] = G2Line
            input_map[f"line_2_{i}"] = G2Line

        for i in range(self.n_fixed_g2, self.n_pairs):
            input_map[f"original_Q{i}"] = G2PointCircuit
            input_map[f"yInv_{i}"] = u384
            input_map[f"xNegOverY_{i}"] = u384
            input_map[f"Q_{i}"] = G2PointCircuit

        input_map.update(
            {
                "R_n_minus_2_of_z": u384,
                "R_n_minus_1_of_z": u384,
                "c_n_minus_3": u384,
                "w_of_z": u384,
                "z": u384,
                "c_inv_frob_1_of_z": u384,
                "c_frob_2_of_z": u384,
                "c_inv_frob_3_of_z": u384,
                "previous_lhs": u384,
                "R_n_minus_3_of_z": u384,
                "Q": (structs.u384Array, self.max_q_degree + 1),
            }
        )
        return input_map

    def _execute_circuit_logic(
        self, circuit: multi_pairing_check.MultiPairingCheckCircuit, vars
    ):
        if self.curve_id == BLS12_381_ID:
            return circuit

        n_pairs = self.n_pairs
        current_points = []
        for k in range(n_pairs):
            circuit.Q.append(
                split_4_sized_object_into_tuple_of_2_size(
                    vars.get(f"original_Q{k}", None)
                )
            )
            circuit.yInv.append(vars[f"yInv_{k}"])
            circuit.xNegOverY.append(vars[f"xNegOverY_{k}"])
            current_points.append(
                split_4_sized_object_into_tuple_of_2_size(vars.get(f"Q_{k}", None))
            )

        circuit.create_lines_z_powers(vars["z"], add_extf_power=True)

        c_n_minus_2 = circuit.square(vars["c_n_minus_3"])
        c_n_minus_1 = circuit.square(c_n_minus_2)

        R_n_minus_2_of_z = vars["R_n_minus_2_of_z"]
        R_n_minus_1_of_z = vars["R_n_minus_1_of_z"]

        # Relation n-2 : f * lines
        prod_k_P_of_z_n_minus_2 = vars["R_n_minus_3_of_z"]  # Init

        lines = circuit.bn254_finalize_step(current_points)
        for l in lines:
            prod_k_P_of_z_n_minus_2 = circuit.mul(
                prod_k_P_of_z_n_minus_2,
                circuit.eval_poly_in_precomputed_Z(
                    l, circuit.line_sparsity, f"line_{k}"
                ),
            )

        lhs_n_minus_2 = circuit.mul(
            c_n_minus_2,
            circuit.sub(prod_k_P_of_z_n_minus_2, R_n_minus_2_of_z),
            comment="c_n_minus_2 * ((Π(n-2,k) (Pk(z)) - R_n_minus_2(z))",
        )

        # Relation n-1 (last one) : f * w * c_inv_frob_1 * c_frob_2 * c_inv_frob_3
        prod_k_P_of_z_n_minus_1 = circuit.mul(
            R_n_minus_2_of_z, vars["c_inv_frob_1_of_z"]
        )
        prod_k_P_of_z_n_minus_1 = circuit.mul(
            prod_k_P_of_z_n_minus_1, vars["c_frob_2_of_z"]
        )
        prod_k_P_of_z_n_minus_1 = circuit.mul(
            prod_k_P_of_z_n_minus_1, vars["c_inv_frob_3_of_z"]
        )
        prod_k_P_of_z_n_minus_1 = circuit.mul(prod_k_P_of_z_n_minus_1, vars["w_of_z"])

        lhs_n_minus_1 = circuit.mul(
            c_n_minus_1,
            circuit.sub(prod_k_P_of_z_n_minus_1, R_n_minus_1_of_z),
            comment="c_n_minus_1 * ((Π(n-1,k) (Pk(z)) - R_n_minus_1(z))",
        )

        _final_lhs = circuit.add(vars["previous_lhs"], lhs_n_minus_2)
        final_lhs = circuit.add(_final_lhs, lhs_n_minus_1)

        Q_of_z = circuit.eval_horner(vars["Q"], z=vars["z"], poly_name="big_Q")
        P_irr, P_irr_sparsity = circuit.write_sparse_constant_elements(
            get_irreducible_poly(self.curve_id, 12).get_coeffs(),
        )
        P_of_z = circuit.eval_poly_in_precomputed_Z(
            P_irr, P_irr_sparsity, poly_name="P_irr"
        )
        check = circuit.sub(final_lhs, circuit.mul(Q_of_z, P_of_z))

        circuit.extend_struct_output(u384("final_check", elmts=[check]))
        return circuit


class MPCheckFinalizeBLS(BaseFixedG2PointsMPCheck):
    def __init__(
        self,
        curve_id: int,
        auto_run: bool = True,
        compilation_mode: int = 1,
        n_pairs: int = 3,
    ):
        self.max_q_degree = multi_pairing_check.get_max_Q_degree(curve_id, n_pairs)
        super().__init__(
            name=f"mp_check_finalize_bls_{n_pairs}P",
            curve_id=curve_id,
            n_pairs=n_pairs,
            n_fixed_g2=None,
            auto_run=auto_run,
            compilation_mode=compilation_mode,
        )

    @property
    def input_map(self):
        return {
            "R_n_minus_1_of_z": u384,
            "c_n_minus_2": u384,
            "w_of_z": u384,
            "z": u384,
            "c_inv_frob_1_of_z": u384,
            "previous_lhs": u384,
            "R_n_minus_2_of_z": u384,
            "Q": (u384Array, self.max_q_degree + 1),
        }

    def _execute_circuit_logic(self, circuit, vars) -> ModuloCircuit:
        if self.curve_id == BN254_ID:
            return circuit

        circuit.create_powers_of_Z(vars["z"], max_degree=12)

        c_n_minus_1 = circuit.square(vars["c_n_minus_2"])
        R_n_minus_1_of_z = vars["R_n_minus_1_of_z"]

        # Relation n-1 (last one) : f * w * c_inv_frob_1
        prod_k_P_of_z_n_minus_1 = circuit.mul(
            vars["R_n_minus_2_of_z"], vars["c_inv_frob_1_of_z"]
        )
        prod_k_P_of_z_n_minus_1 = circuit.mul(prod_k_P_of_z_n_minus_1, vars["w_of_z"])

        lhs_n_minus_1 = circuit.mul(
            c_n_minus_1,
            circuit.sub(prod_k_P_of_z_n_minus_1, R_n_minus_1_of_z),
            comment="c_n_minus_1 * ((Π(n-1,k) (Pk(z)) - R_n_minus_1(z))",
        )

        final_lhs = circuit.add(
            vars["previous_lhs"], lhs_n_minus_1, comment="previous_lhs + lhs_n_minus_1"
        )
        P_irr, P_irr_sparsity = circuit.write_sparse_constant_elements(
            get_irreducible_poly(self.curve_id, 12).get_coeffs()
        )
        P_of_z = circuit.eval_poly_in_precomputed_Z(
            P_irr, P_irr_sparsity, poly_name="P_irr"
        )
        Q_of_z = circuit.eval_horner(vars["Q"], vars["z"], poly_name="big_Q")

        check = circuit.sub(
            final_lhs,
            circuit.mul(Q_of_z, P_of_z, comment="Q(z) * P(z)"),
            comment="final_lhs - Q(z) * P(z)",
        )

        circuit.extend_struct_output(u384("final_check", elmts=[check]))
        return circuit


class MPCheckPreparePairs(BaseFixedG2PointsMPCheck):
    """
    This circuit is used to prepare points for the multi-pairing check.
    For BN curve, it will compute yInv and xNegOverY for each point + negate the y of the G2 point.
    For BLS curve, it will only compute yInv and xNegOverY for each point.
    """

    def __init__(
        self,
        curve_id: int,
        n_pairs: int = 0,
        auto_run: bool = True,
        compilation_mode: int = 1,
    ):
        assert compilation_mode == 1, "Compilation mode 1 is required for this circuit"
        self.n_pairs = n_pairs
        super().__init__(
            name=f"mp_check_prepare_pairs_{n_pairs}P",
            curve_id=curve_id,
            n_pairs=n_pairs,
            n_fixed_g2=None,
            auto_run=auto_run,
            compilation_mode=compilation_mode,
        )
        self.generic_over_curve = True

    @property
    def input_map(self):
        input_map = {}
        for i in range(self.n_pairs):
            input_map[f"p_{i}"] = G1PointCircuit
            if self.curve_id == BN254_ID:
                input_map[f"Qy0_{i}"] = u384
                input_map[f"Qy1_{i}"] = u384
        return input_map

    def _initialize_circuit(self):
        return multi_pairing_check.MultiMillerLoopCircuit(
            self.name,
            self.curve_id,
            n_pairs=self.n_pairs,
            compilation_mode=self.compilation_mode,
        )

    def _execute_circuit_logic(self, circuit, vars) -> ModuloCircuit:
        n_pairs = self.n_pairs
        for i in range(n_pairs):
            p = vars[f"p_{i}"]
            x, y = p
            yInv = circuit.inv(y)
            xNegOverY = circuit.neg(circuit.mul(x, yInv))

            if self.curve_id == BN254_ID:
                Qy0 = vars[f"Qy0_{i}"]
                Qy1 = vars[f"Qy1_{i}"]
                Qyneg0 = circuit.neg(Qy0)
                Qyneg1 = circuit.neg(Qy1)
                circuit.extend_struct_output(
                    BNProcessedPair(
                        name=f"p_{i}", elmts=[yInv, xNegOverY, Qyneg0, Qyneg1]
                    )
                )
            else:
                circuit.extend_struct_output(
                    BLSProcessedPair(name=f"p_{i}", elmts=[yInv, xNegOverY])
                )

        return circuit


class MPCheckPrepareLambdaRootEvaluations(BaseFixedG2PointsMPCheck):
    def __init__(self, curve_id: int, auto_run: bool = True, compilation_mode: int = 1):
        assert compilation_mode == 1, "Compilation mode 1 is required for this circuit"
        super().__init__(
            name="mp_check_prepare_lambda_root",
            curve_id=curve_id,
            n_pairs=2,  # Mocked value, not used in practice.
            n_fixed_g2=None,
            auto_run=auto_run,
            compilation_mode=compilation_mode,
        )

    @property
    def input_map(self):
        input_map = {
            "lambda_root" if self.curve_id == BN254_ID else "lambda_root_inverse": E12D,
            "z": u384,
            "scaling_factor": MillerLoopResultScalingFactor,
        }
        if self.curve_id == BN254_ID:
            input_map.update(
                {
                    "c_inv": E12D,
                    "c_0": u384,
                }
            )
        return input_map

    def _execute_circuit_logic(
        self, circuit: multi_pairing_check.MultiPairingCheckCircuit, vars
    ):
        circuit.create_powers_of_Z(Z=vars["z"], max_degree=11)

        c_or_c_inv = vars[
            "lambda_root" if self.curve_id == BN254_ID else "lambda_root_inverse"
        ]

        if self.curve_id == BLS12_381_ID:
            # Conjugate c_inverse for BLS.
            c_or_c_inv = circuit.conjugate_e12d(c_or_c_inv)

        c_or_c_inv_of_z = circuit.eval_poly_in_precomputed_Z(
            c_or_c_inv, poly_name="C_inv" if self.curve_id == BLS12_381_ID else "C"
        )

        circuit.extend_struct_output(
            u384(
                name="c_inv_of_z" if self.curve_id == BLS12_381_ID else "c_of_z",
                elmts=[c_or_c_inv_of_z],
            )
        )

        scaling_factor_sparsity = [1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0]
        scaling_factor_compressed = vars["scaling_factor"]
        scaling_factor = [
            scaling_factor_compressed[0],
            None,
            scaling_factor_compressed[1],
            None,
            scaling_factor_compressed[2],
            None,
            scaling_factor_compressed[3],
            None,
            scaling_factor_compressed[4],
            None,
            scaling_factor_compressed[5],
            None,
        ]

        scaling_factor_of_z = circuit.eval_poly_in_precomputed_Z(
            scaling_factor, sparsity=scaling_factor_sparsity, poly_name="W"
        )

        circuit.extend_struct_output(
            u384("scaling_factor_of_z", elmts=[scaling_factor_of_z])
        )

        if self.curve_id == BLS12_381_ID:
            c_inv = c_or_c_inv
            c_inv_frob_1 = circuit.frobenius(c_inv, 1)
            c_inv_frob_1_of_z = circuit.eval_poly_in_precomputed_Z(
                c_inv_frob_1, poly_name="C_inv_frob_1"
            )
            circuit.extend_struct_output(
                u384("c_inv_frob_1_of_z", elmts=[c_inv_frob_1_of_z])
            )
        elif self.curve_id == BN254_ID:
            c = c_or_c_inv
            c_of_z = c_or_c_inv_of_z
            c_inv = vars["c_inv"]
            c_0 = vars["c_0"]
            c_inv_of_z = circuit.eval_poly_in_precomputed_Z(c_inv, poly_name="C_inv")
            circuit.extend_struct_output(u384(name="c_inv_of_z", elmts=[c_inv_of_z]))
            lhs = circuit.sub(
                circuit.mul(c_of_z, c_inv_of_z),
                circuit.set_or_get_constant(1),
                comment="c_of_z * c_inv_of_z - 1",
            )
            lhs = circuit.mul(lhs, c_0, comment="c_0 * (c_of_z * c_inv_of_z - 1)")
            circuit.extend_struct_output(u384("lhs", elmts=[lhs]))

            c_inv_frob_1 = circuit.frobenius(c_inv, 1)
            c_frob_2 = circuit.frobenius(c, 2)
            c_inv_frob_3 = circuit.frobenius(c_inv, 3)

            c_inv_frob_1_of_z = circuit.eval_poly_in_precomputed_Z(
                c_inv_frob_1, poly_name="C_inv_frob_1"
            )
            c_frob_2_of_z = circuit.eval_poly_in_precomputed_Z(
                c_frob_2, poly_name="C_frob_2"
            )
            c_inv_frob_3_of_z = circuit.eval_poly_in_precomputed_Z(
                c_inv_frob_3, poly_name="C_inv_frob_3"
            )

            circuit.extend_struct_output(
                u384("c_inv_frob_1_of_z", elmts=[c_inv_frob_1_of_z])
            )
            circuit.extend_struct_output(u384("c_frob_2_of_z", elmts=[c_frob_2_of_z]))
            circuit.extend_struct_output(
                u384("c_inv_frob_3_of_z", elmts=[c_inv_frob_3_of_z])
            )

        return circuit


class FP12MulAssertOne(BaseEXTFCircuit):
    def __init__(
        self,
        curve_id: int,
        auto_run: bool = True,
        init_hash: int = None,
        compilation_mode: int = 0,
    ):
        super().__init__(
            "fp12_mul_assert_one", curve_id, auto_run, init_hash, compilation_mode
        )

    def build_input(self) -> list[PyFelt]:
        input = []
        input.extend([self.field.random() for _ in range(12)])  # X
        input.extend([self.field.random() for _ in range(12)])  # Y
        input.extend([self.field.random() for _ in range(11)])  # Q
        # R is known to be 1.
        input.append(self.field.random())  # z

        return input

    def _run_circuit_inner(self, input: list[PyFelt]) -> ExtensionFieldModuloCircuit:
        circuit = ExtensionFieldModuloCircuit(
            self.name,
            self.curve_id,
            extension_degree=12,
            init_hash=self.init_hash,
            compilation_mode=self.compilation_mode,
        )
        X = circuit.write_struct(E12D("X", elmts=[input.pop(0) for _ in range(12)]))
        Y = circuit.write_struct(E12D("Y", elmts=[input.pop(0) for _ in range(12)]))
        Q = circuit.write_struct(
            structs.E12DMulQuotient("Q", elmts=[input.pop(0) for _ in range(11)])
        )
        assert len(input) == 1
        z = circuit.write_struct(u384("z", elmts=[input.pop(0)]))
        assert len(input) == 0
        circuit.create_powers_of_Z(z, max_degree=12)
        P_irr, P_irr_sparsity = circuit.write_sparse_constant_elements(
            get_irreducible_poly(self.curve_id, 12).get_coeffs(),
        )
        P_of_z = circuit.eval_poly_in_precomputed_Z(
            P_irr, P_irr_sparsity, poly_name="P_irr"
        )
        Q_of_z = circuit.eval_poly_in_precomputed_Z(Q, poly_name="Q")
        R_of_z = circuit.set_or_get_constant(1)

        X_of_z = circuit.eval_poly_in_precomputed_Z(X, poly_name="X")
        Y_of_z = circuit.eval_poly_in_precomputed_Z(Y, poly_name="Y")
        check = circuit.sub(
            circuit.mul(X_of_z, Y_of_z, comment="X(z) * Y(z)"),
            circuit.mul(Q_of_z, P_of_z, comment="Q(z) * P(z)"),
            comment="(X(z) * Y(z)) - (Q(z) * P(z))",
        )
        check = circuit.sub(check, R_of_z, comment="(X(z) * Y(z) - Q(z) * P(z)) - 1")
        circuit.extend_struct_output(u384("check", elmts=[check]))

        return circuit


class EvalE12D(BaseEXTFCircuit):
    def __init__(
        self,
        curve_id: int,
        auto_run: bool = True,
        init_hash: int = None,
        compilation_mode: int = 0,
    ):
        super().__init__("eval_e12d", curve_id, auto_run, init_hash, compilation_mode)

    def build_input(self) -> list[PyFelt]:
        input = []
        input.extend([self.field.random() for _ in range(12)])  # X
        input.append(self.field.random())  # z

        return input

    def _run_circuit_inner(self, input: list[PyFelt]) -> ExtensionFieldModuloCircuit:
        circuit = ExtensionFieldModuloCircuit(
            self.name,
            self.curve_id,
            extension_degree=12,
            init_hash=self.init_hash,
            compilation_mode=self.compilation_mode,
        )
        X = circuit.write_struct(E12D("f", elmts=[input.pop(0) for _ in range(12)]))
        z = circuit.write_struct(u384("z", elmts=[input.pop(0)]))
        X_of_z = circuit.eval_horner(X, z, poly_name="X")
        circuit.extend_struct_output(u384("f_of_z", elmts=[X_of_z]))
        return circuit



import math

from garaga import garaga_rs
from garaga.algebra import PyFelt
from garaga.definitions import CURVES, CurveID, G1G2Pair, G1Point, G2Point
from garaga.hints.bls import get_root_and_scaling_factor_bls
from garaga.hints.tower_backup import E12


def get_final_exp_witness(
    curve_id: int, f: E12, use_rust: bool = True
) -> tuple[E12, E12]:
    """
    From a miller loop output f of a given curve ID such that f^h == 1,
    Returns the witnesses c, w for the final exponentiation step, such that
    - f*w = c^lambda
    - w lies in a subfield of Fq12.
    """
    if curve_id != CurveID.BN254.value and curve_id != CurveID.BLS12_381.value:
        raise ValueError(f"Curve ID {curve_id} not supported")

    curve = CURVES[curve_id]
    f_values = f.value_coeffs

    if use_rust:
        c_values, wi_values = garaga_rs.get_final_exp_witness(curve_id, f_values)
        c = E12([PyFelt(v, curve.p) for v in c_values], curve_id)
        wi = E12([PyFelt(v, curve.p) for v in wi_values], curve_id)
    else:
        if curve_id == CurveID.BN254.value:
            c, wi = find_c_e12(f, get_27th_bn254_root())
        elif curve_id == CurveID.BLS12_381.value:
            c, wi = get_root_and_scaling_factor_bls(f)
        else:
            raise ValueError(f"Curve ID {curve_id} not supported")

    return c, wi


def get_lambda(curve_id: CurveID) -> int:
    x = CURVES[curve_id.value].x
    q = CURVES[curve_id.value].p
    if curve_id == CurveID.BN254:
        λ = (
            6 * x + 2 + q - q**2 + q**3
        )  # https://eprint.iacr.org/2008/096.pdf See section 4 for BN curves.
        return λ
    elif curve_id == CurveID.BLS12_381:
        λ = -x + q
        return λ
    else:
        raise ValueError(f"Curve ID {curve_id} not supported")


def get_m_dash_root(f: E12) -> E12:
    assert f.curve_id == CurveID.BN254.value

    q = CURVES[f.curve_id].p
    x = CURVES[f.curve_id].x
    r = CURVES[f.curve_id].n

    h = (q**12 - 1) // r  # = 3^3 · l # where gcd(l, 3) = 1
    assert math.gcd(r, h) == 1
    base = 3
    k, l = decompose_scalar_into_b_powers_and_remainder(h, base)
    assert base**k * l == h, f"{base}^k * l should be h"
    assert h % (base**k) == 0, f"h should be a multiple of {base}^k"
    assert math.gcd(l, base) == 1, f"l should be coprime with {base}"

    λ = (
        6 * x + 2 + q - q**2 + q**3
    )  # https://eprint.iacr.org/2008/096.pdf See section 4 for BN curves.

    assert λ % r == 0, "λ should be a multiple of r. See section 4.2.2"
    m = λ // r
    d = math.gcd(m, h)

    assert m % d == 0, "m should be a multiple of d"
    m_dash = m // d  # m' = m/d
    assert m_dash % h != 0
    "m/d should not divide h. See section 4.2.2 Theorem 2."
    assert d * r * m_dash == λ, "incorrect parameters"  # sanity check
    assert math.gcd(λ, q**12 - 1) == d * r
    assert (
        math.gcd(m_dash, q**12 - 1) == 1
    ), "m_dash should be coprime with q**12 - 1 'by construction'. See 4.3.2 computing m-th root"
    m_d_inv = pow(m_dash, -1, h)

    return f**m_d_inv


def decompose_scalar_into_b_powers_and_remainder(scalar: int, b: int):
    """
    Decompose scalar into b^k * l, where l is not divisible by b.
    """

    k = 0
    l = scalar

    while l % b == 0:
        l //= b
        k += 1
    assert l % b != 0, "l should not be divisible by b"
    assert scalar == b**k * l, "scalar should be the product of b^k * l"
    return k, l


def pow_3_ord(a: E12):
    t = 0
    FP12_ONE = E12.one(a.curve_id)
    while a != FP12_ONE:
        t += 1
        a = a**3
    return t


def find_cube_root(a: E12, w: E12, q: int) -> E12:
    # Algorithm 4: Modified Tonelli-Shanks for cube roots
    # Input: Cube residue a, cube non residue w and write p − 1 = 3^r · s such that 3 ∤ s
    # Output: x such that x^3 = a
    # 1 exp = (s + 1)/3
    _, s = decompose_scalar_into_b_powers_and_remainder(q**12 - 1, 3)
    exp: int = (s + 1) // 3
    a_inv = a.__inv__()
    # 2 x ← a^exp
    x = a**exp
    # 3 3^t ← ord((x^3)/a)
    t = pow_3_ord(x**3 * a_inv)
    # 4 while t != 0 do
    while t != 0:
        # 5 exp = (s + 1)/3
        # 6 x ← x · w^exp
        x = x * w**exp
        # 7 3^t ← ord(x^3/a)
        t = pow_3_ord(x**3 * a_inv)
    # 8 end
    # 9 return x
    return x


def find_c_e12(f: E12, w: E12) -> tuple[E12, E12]:
    # Algorithm 5: Algorithm for computing λ residues over BN curve
    # Input: Output of a Miller loop f and fixed 27-th root of unity w
    # Output: (c, wi) such that c**λ = f · wi
    # 1 s = 0
    s = 0
    q = CURVES[f.curve_id].p
    exp = (q**12 - 1) // 3
    FP12_ONE = E12.one(f.curve_id)
    # 2 if f**(q**k-1)/3 = 1 then
    if f**exp == FP12_ONE:
        # 3 continue
        # 4 end
        # 5 else if (f · w)**(q**k-1)/3 = 1 then
        c = f
    elif (f * w) ** exp == FP12_ONE:
        # 6 s = 1
        s = 1
        # 7 f ← f · w
        c = f * w
    # 8 end
    # 9 else
    else:
        # 10 s = 2
        s = 2
        # 11 f ← f · w**2
        c = f * w * w
    # 12 end
    # 13 c ← f**r′
    c = get_rth_root(c)
    # 14 c ← c**m′′
    c = get_m_dash_root(c)
    # 15 c ← c**1/3 (by using modified Tonelli-Shanks 4)
    c = find_cube_root(c, w, q)
    # 16 return (c, ws)
    return c, w**s


def get_rth_root(f: E12) -> E12:
    """
    Computes x such that x^r = f
    """
    r = CURVES[f.curve_id].n
    h = (CURVES[f.curve_id].p ** 12 - 1) // r
    r_inv = pow(r, -1, h)
    res = f**r_inv
    # assert res**r == f, "res**r should be f"
    return res


def get_27th_bn254_root():
    """
    Retrieve a 27th root of unity over BN254 Fp^12 that isn't a 9th root of unity.
    """
    root_27th = E12(
        [
            0,
            0,
            0,
            0,
            8204864362109909869166472767738877274689483185363591877943943203703805152849,
            17912368812864921115467448876996876278487602260484145953989158612875588124088,
            0,
            0,
            0,
            0,
            0,
            0,
        ],
        CurveID.BN254.value,
    )
    FP12_ONE = E12.one(CurveID.BN254.value)
    assert root_27th**27 == FP12_ONE, "root_27th**27 should be one"
    assert root_27th**9 != FP12_ONE, "root_27th**9 should not be one"
    return root_27th


def get_miller_loop_output(curve_id: CurveID, will_be_one: bool = True) -> E12:
    """
    Returns a random miller loop output f such that f**h = 1
    """
    g1, g2 = G1Point.gen_random_point(curve_id), G2Point.gen_random_point(curve_id)
    if will_be_one:
        # Miller (-g1, g2) * Miller (g1, g2)
        f = G1G2Pair.miller([G1G2Pair(-g1, g2), G1G2Pair(g1, g2)])
        h = (CURVES[curve_id.value].p ** 12 - 1) // CURVES[curve_id.value].n
        assert f**h == E12.one(curve_id.value), "f**h should be one"
    else:
        f = G1G2Pair.miller([G1G2Pair(g1, g2)])

    return f


if __name__ == "__main__":

    def test_bn254():
        f = get_miller_loop_output(CurveID.BN254)
        x = CURVES[CurveID.BN254.value].x
        q = CURVES[CurveID.BN254.value].p
        λ = (
            6 * x + 2 + q - q**2 + q**3
        )  # https://eprint.iacr.org/2008/096.pdf See section 4 for BN curves.

        c, wi = find_c_e12(f, get_27th_bn254_root())
        print(f"wi: {wi.value_coeffs}")
        c_inv = c.__inv__()
        result = c_inv**λ * f * wi
        assert result == E12.one(CurveID.BN254.value), "pairing not 1"

    for i in range(10):
        test_bn254()
        print(f"Test {i} passed")
